"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import argparse
import bottle as bot  # type: ignore
import ifpd
from ifpd.scripts import arguments as ap  # type: ignore
from ifpd.exception import enable_rich_assert
import importlib.util
import logging
import os
from rich.logging import RichHandler  # type: ignore

logging.basicConfig(
    level=logging.INFO,
    format="%(message)s",
    handlers=[RichHandler(markup=True, rich_tracebacks=True)],
)


def init_parser(subparsers: argparse._SubParsersAction) -> argparse.ArgumentParser:
    parser = subparsers.add_parser(
        __name__.split(".")[-1],
        description="Run WebServer.",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        help="Run WebServer.",
    )
    parser.add_argument(
        "static",
        metavar="folder",
        type=str,
        help="Path to static folder (created if not found).",
    )

    parser.add_argument(
        "-u",
        "--url",
        metavar="url",
        type=str,
        default="0.0.0.0",
        help="URL hosting the web server. Default: 0.0.0.0",
    )
    parser.add_argument(
        "-p",
        "--port",
        metavar="port",
        type=int,
        default=8080,
        help="Web server port. Default: 8080",
    )
    parser.add_argument(
        "-m",
        "--mail",
        metavar="email",
        type=str,
        default="email@example.com",
        help="Email address of server admin.",
    )
    parser = ap.add_version_option(parser)

    advanced = parser.add_argument_group("advanced arguments")
    advanced.add_argument(
        "--hide-breadcrumbs",
        action="store_const",
        dest="show_breadcrumbs",
        const=False,
        default=True,
        help="""Hide navigation breadcrumbs.""",
    )
    advanced.add_argument(
        "-R",
        "--custom-routes",
        metavar="routesFile",
        type=str,
        help="Path to custom routes Python file.",
    )
    parser.add_argument(
        "-T",
        "--custom-templates",
        metavar="templateFolder",
        type=str,
        help="Path to folder with custom templates.",
    )
    parser.add_argument(
        "-H",
        "--homepage",
        metavar="homepageTemplate",
        type=str,
        help="""Name of homepage template. Homepage is off by default.
        Use "-H home_default" to turn default homepage template on.
        When using a custom homepage template, -T must be specified.""",
    )

    parser.set_defaults(parse=parse_arguments, run=run)

    return parser


@enable_rich_assert
def parse_arguments(args: argparse.Namespace) -> argparse.Namespace:
    return args


def add_static_routes_includes(root, root_path):
    # CSS files
    @root.route("/css/<path>")
    def callback_css(path):
        return bot.static_file(path, "%s/interface/css/" % root_path)

    # JS files
    @root.route("/js/<path>")
    def callback_js(path):
        return bot.static_file(path, "%s/interface/js/" % root_path)

    # Fonts files
    @root.route("/fonts/<path>")
    def callback_fonts(path):
        return bot.static_file(path, "%s/interface/fonts/" % root_path)

    return root


def add_static_routes_download(root, root_path):
    # Images
    @root.route("/images/<path>")
    def callback_images(path):
        return bot.static_file(path, "%s/interface/images/" % root_path)

    # Documents
    @root.route("/documents/<path>")
    def callback_documents(path):
        return bot.static_file(path, "%s/interface/documents/" % root_path)

    return root


def build_root_app(args, home_template, home_status):
    # Server params
    root_path = "%s/" % os.path.dirname(ifpd.__file__)
    section_path = "%s/" % os.path.dirname(ifpd.sections.__file__)

    # Start root app
    root = bot.Bottle()

    # Home
    @root.route("/")
    @bot.view(home_template)
    def index():
        return {
            "custom_stylesheets": ["home.css"],
            "title": "iFISH",
            "description": "iFISH",
            "home_status": home_status,
        }

    # 404 Error
    @root.error(404)
    def error404(error):
        return "Nothing here, sorry :("

    root = add_static_routes_includes(root, root_path)
    root = add_static_routes_download(root, root_path)

    # Load Sections
    pdApp = ifpd.sections.probe_design.App(
        section_path,
        args.static,
        root_path,
        "http://%s:%d/" % (args.url, args.port),
        "probe-design/",
    )
    pdApp.admin_email = args.mail

    # Custom routes
    if args.custom_routes is not None:
        assert args.custom_routes.endswith(".py")
        spec = importlib.util.spec_from_file_location(
            "custom_routes", args.custom_routes
        )
        custom_routes_mod = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(custom_routes_mod)
        if "update_probe_design_app" in dir(custom_routes_mod):
            pdApp = custom_routes_mod.update_probe_design_app(bot, pdApp, args)
        if "update_root_app" in dir(custom_routes_mod):
            root = custom_routes_mod.update_root_app(bot, root, args)

    pdApp.vd["breadcrumbs"] = args.show_breadcrumbs

    # Mount Sections
    root.mount("probe-design", pdApp)

    return root


def mk_missing_dirs(args):
    if not os.path.isdir(args.static):
        os.mkdir(args.static)
    if not os.path.isdir("%s/db" % args.static):
        os.mkdir("%s/db" % args.static)
    if not os.path.isdir("%s/query" % args.static):
        os.mkdir("%s/query" % args.static)


@enable_rich_assert
def run(args: argparse.Namespace) -> None:
    # Server params
    root_path = "%s/" % os.path.dirname(ifpd.__file__)

    bot.TEMPLATE_PATH.append(root_path)
    bot.TEMPLATE_PATH.append("%s/interface/views/" % root_path)
    if args.custom_templates is not None:
        assert os.path.isdir(
            args.custom_templates
        ), f"folder not found: '{args.custom_templates}'"
        bot.TEMPLATE_PATH.append(args.custom_templates)

    assert not os.path.isfile(args.static), (
        "folder expected, file found: %s" % args.static
    )
    mk_missing_dirs(args)

    if args.homepage is not None:
        if args.homepage != "home_default.tpl.html":
            assert (
                args.custom_templates is not None
            ), "-T option required when using -H."
        home_template = args.homepage
        home_status = True
    else:
        home_template = "home_default.tpl.html"
        home_status = False

    if args.custom_routes is not None:
        assert os.path.isfile(
            args.custom_routes
        ), f"file not found: {args.custom_routes}"

    root = build_root_app(args, home_template, home_status)

    # RUN ==========================================================================

    root.run(host=args.url, port=args.port, debug=True, server="paste")

    logging.info("Done. :thumbs_up: :smiley:")



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

from ifpd.sections import app as rootApp
from ifpd.sections.probe_design.enquirer import Enquirer
from ifpd.sections.probe_design.queue import Queue
from ifpd.sections.probe_design.routes import Routes
import logging
from typing import Dict


class App(rootApp.App):
    """Probe Design Bottle App.

    Args:
            app_uri (string): section relative url.
            base_dir (string): section base directory.
            consumer (Enquirer): queue consumer.
            local_path (string): absolute path to app directory.
            qpath (string): absolute path to query folder.
            queue (Queue): query queue.
            root_path (string): webserver root absolute path.
            root_uri (string): root webserver url.
            route_list (string): probe_design.routes.Routes instance.
            section_path (string): sections relative path.
            tprefix (string): title prefix.
            vd (string): view data.
            vpath (string): absolute path to views folder.
            BUF_SIZE (int): queue size.
    """

    vd: Dict = {}
    route_list = Routes()
    base_dir = "probe_design"
    tprefix = "Probe Designer ~ "
    vd = {}
    BUF_SIZE = 0
    MAX_CURR = 1

    def __init__(
        self,
        section_path,
        static_path,
        root_path,
        root_uri,
        app_uri,
        MAX_CURR=None,
        BUF_SIZE=None,
    ):
        """Initialize.

        Args:
                section_path (string): sections relative path.
                root_path (string): webserver root absolute path.
                root_uri (string): root webserver url.
                app_uri (string): section relative url.
                MAX_CURR (int): maximum number of running tasks.
                BUF_SIZE (int): queue size, defaults to 0.
        """

        # Run default initialization
        super(App, self).__init__()

        # Save input parameters
        self.section_path = section_path
        self.root_uri = root_uri
        if root_uri.endswith("/") and root_uri[:-1].endswith(":80"):
            uri_for_check = ":".join(root_uri.split(":")[:-1]) + "/"
        else:
            uri_for_check = root_uri
        self.app_uri = app_uri
        self.local_path = "%s%s" % (section_path, self.base_dir)
        self.static_path = static_path
        self.vpath = "%s/views/" % self.local_path
        self.qpath = "%s/query/" % self.static_path
        self.MAX_CURR = MAX_CURR
        self.BUF_SIZE = BUF_SIZE

        # Start setting up view dictionary
        self.vd["app_uri"] = "%s%s" % (root_uri, app_uri)
        self.vd["root_uri"] = root_uri
        self.vd["uri_for_check"] = uri_for_check
        self.vd["vpath"] = self.vpath
        self.vd["qpath"] = self.qpath
        self.vd["description"] = "iFISH probe designer application."
        self.vd["breadcrumbs"] = False
        self.vd["menu_template"] = None
        self.vd["SHOW_COOKIE_CONSENT_BANNER"] = False
        self.vd["google_analytics_token"] = ""

        # Set default
        if type(None) != type(MAX_CURR):
            self.MAX_CURR = MAX_CURR
        if type(None) != type(BUF_SIZE):
            self.BUF_SIZE = BUF_SIZE

        # Logging config
        logging.basicConfig(
            level=logging.DEBUG, format="(%(threadName)-9s) %(message)s"
        )

        # Initialize queue
        self.queue = Queue(BUF_SIZE=self.BUF_SIZE, MAX_CURR=self.MAX_CURR)
        self.consumer = Enquirer(self.queue)
        self.consumer.start()

        # Save queue
        self.vd["queue"] = self.queue

        # Build routes
        self.build_routes()

        # Output
        return



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

# import ifpd as fp
import os
import pandas as pd  # type: ignore
from rich.progress import track  # type: ignore


class UCSCbed(object):
    """Class to read a UCSC Bed format file into a pandas.DataFrame."""

    FIELD_NAMES = [
        "chrom",
        "chromStart",
        "chromEnd",
        "name",
        "score",
        "strand",
        "thickStart",
        "thickEnd",
        "itemRgb",
        "blockCount",
        "blockSizes",
        "blockStarts",
    ]
    FIELD_FORMAT = "siisfsiisiss"
    FORMATTER = {"f": float, "i": int, "s": str}

    def __init__(self, path, incrementChromEnd=False, bufferize=False):
        super(UCSCbed, self).__init__()
        self.path = path
        self.incrementChromEnd = incrementChromEnd
        self.nrecords = self.count_records()
        if not bufferize:
            self.__read()

    def count_records(self):
        return sum(1 for line in open(self.path))

    def __read(self):
        """Reads the bed file into a pandas.DataFrame."""
        assert os.path.isfile(self.path), f"bed file not found: '{self.path}'"

        bedDF = []
        with open(self.path, "r+") as IH:
            header = next(IH)
            if header.startswith("browser") or header.startswith("track"):
                self.custom = True
                self.header = header.strip()
            else:
                bedDF.append(UCSCbed.parse_bed_line(header))
                self.custom = False
                self.header = None
            bedDF.extend([UCSCbed.parse_bed_line(line) for line in IH])

        self.df = pd.concat(bedDF)
        self.df.index = range(self.df.shape[0])
        self.ncols = self.df.shape[1]

    def __set_custom_header(self):
        with open(self.path, "r+") as IH:
            line = next(IH)
            if line.startswith("browser") or line.startswith("track"):
                self.custom = True
                self.header = line.strip()
            else:
                self.custom = False
                self.header = None

    def buffer(self, parse=True, enforceBED3=False):
        """Reads the bed file and yields one line at a time. The content is not
        stored in the class. Each call to the buffer function moves forward to
        the next line. To restart, re-initialize the class with bufferize=True.
        Use enforceBED3 to strip any additional column. Use parse to get a
        formatted pd.DataFrame instead of a raw string."""
        self.__set_custom_header()
        with open(self.path, "r+") as IH:
            if self.custom:
                next(IH)
            for line in track(IH, total=self.nrecords - 1, description="Parsing"):
                if parse:
                    yield UCSCbed.parse_bed_line(
                        line, enforceBED3, self.incrementChromEnd
                    )
                else:
                    if self.incrementChromEnd:
                        line = line.strip().split("\t")
                        line[2] = str(int(line[2]) + 1)
                        line = "\t".join(line) + "\n"
                    yield line

    @staticmethod
    def parse_bed_line(line, enforceBED3=False, incrementChromEnd=False):
        """Parses and checks one line of a bed file.
        Does not work on header lines."""

        line = line.strip().split("\t")
        assert 3 <= len(line), f"at least 3 fields required, {len(line)} found."
        line = line[: min(4 - enforceBED3, len(line))]

        lineDF = pd.DataFrame(
            [
                UCSCbed.FORMATTER[UCSCbed.FIELD_FORMAT[i]](line[i])
                for i in range(len(line))
            ]
        ).transpose()
        lineDF.columns = UCSCbed.FIELD_NAMES[: len(line)]

        lineDF.iloc[0, 2] += incrementChromEnd

        return lineDF

    def isBEDN(self, n):
        n = min(n, 12)
        return self.ncols == n

    def getBEDN(self, n):
        n = min(n, 12)
        assert self.ncols >= n
        return self.df.iloc[:, :n]

    def mkBEDN(self, n):
        assert n <= 12
        if n < self.ncols:
            self.df = self.df.iloc[:, :n]
            self.ncols = n



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import argparse
from ifpd import query
from ifpd.scripts import arguments as ap  # type: ignore
from ifpd.exception import enable_rich_assert
import logging
from rich.logging import RichHandler  # type: ignore

logging.basicConfig(
    level=logging.INFO,
    format="%(message)s",
    handlers=[RichHandler(markup=True, rich_tracebacks=True)],
)


def init_parser(subparsers: argparse._SubParsersAction) -> argparse.ArgumentParser:
    parser = subparsers.add_parser(
        __name__.split(".")[-1],
        description="""
Checks integrity of a database. Specifically:
- a ".config" file must be present
    * Overlapping status must match.
- At least one non-empty chromosome file
    * Appropriate format (3 columns).
    * Positions must be sorted.
    * An ODN must end after it starts.
    * No ODNs can start from the same position.
    * No ODNs can be totally included in another one.
""",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        help="Check integrity of a database.",
    )

    parser.add_argument(
        "dbDirPath", type=str, default=".", help="""Path to database directory."""
    )
    parser = ap.add_version_option(parser)

    parser.set_defaults(parse=parse_arguments, run=run)

    return parser


@enable_rich_assert
def parse_arguments(args: argparse.Namespace) -> argparse.Namespace:
    return args


@enable_rich_assert
def run(args: argparse.Namespace) -> None:
    logging.info("Read database.")
    oligoDB = query.OligoDatabase(args.dbDirPath)
    oligoDB.read_all_chromosomes(True)
    logging.info(f'Database name: "{oligoDB.get_name()}"')

    refGenome = oligoDB.get_reference_genome()
    logging.info(f'Reference genome: "{refGenome}"')

    oligoMinDist = oligoDB.get_oligo_min_dist()
    oligoLengthRange = oligoDB.get_oligo_length_range()
    hasOverlaps = oligoDB.has_overlaps()
    logging.info(f"Contains overlaps: {hasOverlaps}")
    logging.info(f"Oligo min distance: {oligoMinDist}")
    logging.info(f"Oligo length range: {oligoLengthRange}")
    logging.info("Database checked.")

    logging.info("Done. :thumbs_up: :smiley:")



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

from ifpd.scripts import arguments, ifpd
from ifpd.scripts import dbchk, mkdb
from ifpd.scripts import query
from ifpd.scripts import serve

__all__ = ["arguments", "ifpd", "dbchk", "mkdb", "query", "serve"]



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import argparse
from ifpd import __version__
import joblib  # type: ignore
import logging
from rich.logging import RichHandler  # type: ignore
from rich.console import Console  # type: ignore
import os
import sys
from typing import Optional


def add_version_option(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    parser.add_argument(
        "--version", action="version", version=f"{sys.argv[0]} {__version__}"
    )
    return parser


def check_n_oligo(args, selectCondition):
    assert 0 < selectCondition.sum(), "".join(
        [
            "no oligos found in the specified region.",
            f" [{args.chrom}:{args.region[0]}-{args.region[1]}]",
        ]
    )
    if args.exact_n_oligo:
        assert args.n_oligo <= selectCondition.sum(), "".join(
            [
                "there are not enough oligos in the database.",
                f" Asked for {args.n_oligo}, {selectCondition.sum()} found.",
            ]
        )
    elif args.n_oligo > selectCondition.sum():
        logging.info(
            "".join(
                [
                    f"Found {selectCondition.sum()} oligos in",
                    f" {args.chrom}:{args.region[0]}-{args.region[1]}",
                ]
            )
        )
        logging.warning(
            "".join(
                [
                    f"Designing a probe with {selectCondition.sum()} oligos",
                    f" (instead of {args.n_oligo}).",
                ]
            )
        )
        args.n_oligo = selectCondition.sum()
    return args


def add_log_file_handler(path: str, logger_name: Optional[str] = None) -> None:
    """Adds log file handler to logger.

    By defaults, adds the handler to the root logger.

    Arguments:
        path {str} -- path to output log file

    Keyword Arguments:
        logger_name {str} -- logger name (default: {""})
    """
    assert not os.path.isdir(path)
    log_dir = os.path.dirname(path)
    assert os.path.isdir(log_dir) or log_dir == ""
    fh = RichHandler(console=Console(file=open(path, mode="w+")), markup=True)
    fh.setLevel(logging.INFO)
    logging.getLogger(logger_name).addHandler(fh)
    logging.info(f"[green]Log to[/]: '{path}'")


def check_threads(threads: int) -> int:
    if threads > joblib.cpu_count():
        return joblib.cpu_count()
    elif threads <= 0:
        return 1
    return threads



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import matplotlib  # type: ignore
import matplotlib.pyplot as plt  # type: ignore
import matplotlib.patches as patches  # type: ignore
import configparser
from ifpd import bioext, stats
from joblib import Parallel, delayed  # type: ignore
import numpy as np  # type: ignore
import os
import pandas as pd  # type: ignore
from rich.progress import track  # type: ignore
from typing import List

matplotlib.use("svg")


class OligoDatabase(object):
    """FISH-ProDe Oligonucleotide Database class."""

    def __init__(self, dbDirPath):
        super(OligoDatabase, self).__init__()
        self.dirPath = dbDirPath
        self.chromData = {}

        assert not os.path.isfile(
            self.dirPath
        ), f'expected folder, file found: "{self.dirPath}"'
        assert os.path.isdir(
            self.dirPath
        ), f'database folder not found: "{self.dirPath}"'

        configPath = os.path.join(self.dirPath, ".config")
        assert os.path.isfile(configPath), f'missing "{configPath}" file.'

        with open(configPath, "r") as IH:
            self.config = configparser.ConfigParser()
            self.config.read_string("".join(IH.readlines()))

    def check_overlaps(self):
        hasOverlaps = False
        for chrom, chromData in self.chromData.items():
            startPositions = np.array(chromData.iloc[:-1, 0])
            endPositions = np.array(chromData.iloc[1:, 1]) - 1
            foundOverlaps = any(startPositions <= endPositions)
            hasOverlaps |= foundOverlaps
        return hasOverlaps == self.has_overlaps()

    def get_oligo_length_range(self):
        """Reads oligo length range from Database .config"""
        return (
            self.config.getint("OLIGOS", "min_length"),
            self.config.getint("OLIGOS", "max_length"),
        )

    def get_oligo_min_dist(self):
        """Reads consecutive oligo minimum distance from Database .config"""
        return self.config.getint("OLIGOS", "min_dist")

    def get_name(self):
        """Reads Database name from Database .config"""
        return self.config["DATABASE"]["name"]

    def get_reference_genome(self):
        """Reads reference genome from Database .config"""
        return self.config["DATABASE"]["refGenome"]

    def has_overlaps(self):
        """Reads overlaps status from Database .config"""
        return self.config.getboolean("OLIGOS", "overlaps")

    def has_sequences(self):
        """Reads sequence status from Database .config"""
        return True

    def has_chromosome(self, chrom):
        return chrom in os.listdir(self.dirPath)

    def read_chromosome(self, chrom):
        assert self.has_chromosome(chrom)

        chromPath = os.path.join(self.dirPath, chrom)
        chromData = pd.read_csv(chromPath, "\t", header=None)
        chromData.columns = bioext.UCSCbed.FIELD_NAMES[1 : (chromData.shape[1] + 1)]

        assert 0 != chromData.shape[0], f'found empty chromosome file: "{chromPath}"'
        assert 2 <= chromData.shape[1], f'missing columns in "{chromPath}"'
        assert all(
            chromData.iloc[:, 0].diff()[1:] > 0
        ), f'found unsorted file: "{chromPath}"'
        assert all(
            chromData.iloc[:, 1].diff()[1:] >= 0
        ), 'found unsorted file: "{chromPath}"'

        oligoMinDist = self.get_oligo_min_dist()
        startValues = chromData.iloc[1:, 0].values
        endValues = chromData.iloc[:-1, 1].values
        if len(startValues) != 0:
            oligoMinD_observed = min(startValues - endValues)
        else:
            oligoMinD_observed = np.inf
        assert oligoMinD_observed >= oligoMinDist, "".join(
            [
                "oligo min distance does not match: ",
                f"{oligoMinD_observed} instead of {oligoMinDist}.",
            ]
        )

        oligoLengthRange = self.get_oligo_length_range()
        oligoLengthList = chromData.iloc[:, 1] - chromData.iloc[:, 0]
        assert all(
            oligoLengthList >= oligoLengthRange[0]
        ), f'oligo too small for ".config" in "{chromPath}"'
        assert all(
            oligoLengthList <= oligoLengthRange[1]
        ), f'oligo too big for ".config" in "{chromPath}"'

        if self.has_overlaps():
            assert self.check_overlaps(), f'overlaps status mismatch in "{chromPath}"'

        if self.has_sequences():
            assert chromData.shape[1] >= 3, f'missing sequence columns in "{chromPath}"'

        self.chromData[chrom] = chromData

    def read_all_chromosomes(self, verbose):
        chromList = [d for d in os.listdir(self.dirPath) if not d.startswith(".")]
        assert 0 < len(chromList), "no chromosome files found in {self.dirPath}"
        chromList = track(chromList) if verbose else chromList

        for chrom in chromList:
            self.read_chromosome(chrom)


class OligoProbe(object):
    """Class for probe management."""

    def __init__(self, chrom, oligos, database):
        super(OligoProbe, self).__init__()
        self.chrom = chrom
        self.oligoData = oligos
        self.refGenome = database.get_reference_genome()
        self.chromStart = self.oligoData.iloc[:, 0].min()
        self.chromEnd = self.oligoData.iloc[:, 1].max()
        self.midpoint = (self.chromStart + self.chromEnd) / 2
        self.size = self.chromEnd - self.chromStart
        self.homogeneity = self.get_probe_homogeneity()

    def __str__(self):
        s = f"[{self.refGenome}]"
        s += f"{self.chrom}:{self.chromStart}-{self.chromEnd};"
        s += f" oligoSpread: {self.homogeneity}"
        return s

    def asDataFrame(self, region=None):
        if type(None) == type(region):
            return pd.DataFrame.from_dict(
                {
                    "chrom": [self.chrom],
                    "chromStart": [self.chromStart],
                    "chromEnd": [self.chromEnd],
                    "refGenome": [self.refGenome],
                    "midpoint": [self.midpoint],
                    "size": [self.size],
                    "homogeneity": [self.homogeneity],
                }
            )
        else:
            return pd.DataFrame.from_dict(
                {
                    "chrom": [self.chrom],
                    "chromStart": [self.chromStart],
                    "chromEnd": [self.chromEnd],
                    "refGenome": [self.refGenome],
                    "midpoint": [self.midpoint],
                    "size": [self.size],
                    "homogeneity": [self.homogeneity],
                    "regChromStart": region[0],
                    "regChromEnd": region[1],
                }
            )

    def get_probe_centrality(self, region):
        """Calculate centrality, as location of the probe midpoint relative to
        the region midpoint. 1 when overlapping, 0 when the probe midpoint is
        at either of the region extremities."""
        region_halfWidth = (region[2] - region[1]) / 2
        region_midPoint = region[1] + region_halfWidth
        return (
            region_halfWidth - abs(region_midPoint - self.midpoint)
        ) / region_halfWidth

    def get_probe_size(self):
        """Probe size, defined as difference between start of first oligo
        and end of the last one."""
        return self.oligoData.iloc[:, 1].max() - self.oligoData.iloc[:, 0].min()

    def get_probe_homogeneity(self):
        """Probe homogeneity, as the inverse of the standard deviation of the
        distance between consecutive oligos, calculated from their start
        position, disregarding oligo length."""
        std = np.std(np.diff(self.oligoData.iloc[:, 1]))
        return np.inf if std == 0 else 1 / std

    def describe(self, region, path=None):
        """Builds a small pd.DataFrame describing a probe."""
        description = pd.DataFrame.from_dict(
            {
                "chrom": [region[0]],
                "chromStart": [self.oligoData.iloc[:, 0].min()],
                "chromEnd": [self.oligoData.iloc[:, 1].max()],
                "centrality": [self.get_probe_centrality(region)],
                "size": [self.size],
                "homogeneity": [self.homogeneity],
            }
        )

        if type(None) != type(path):
            config = configparser.ConfigParser()
            config["REGION"] = {
                "chrom": region[0],
                "chromStart": region[1],
                "chromEnd": region[2],
            }
            config["PROBE"] = {
                "chrom": description["chrom"].values[0],
                "chromStart": description["chromStart"].values[0],
                "chromEnd": description["chromEnd"].values[0],
                "nOligo": self.oligoData.shape[0],
            }
            config["FEATURES"] = {
                "centrality": description["centrality"].values[0],
                "size": description["size"].values[0],
                "homogeneity": description["homogeneity"].values[0],
            }
            with open(path, "w+") as OH:
                config.write(OH)

        return description

    def get_fasta(self, path=None, prefix=""):
        if not prefix.startswith(" "):
            prefix = " " + prefix

        fasta = ""
        for i in self.oligoData.index:
            oligo = self.oligoData.loc[i, :]
            chromStart, chromEnd, sequence = oligo[:3]
            fasta += f">{prefix}oligo_{i} [{self.refGenome}]"
            fasta += f"{self.chrom}:{chromStart}-{chromEnd}\n"
            fasta += f"{sequence}\n"

        if type(None) != type(path):
            assert os.path.isdir(os.path.dirname(path))
            with open(path, "w+") as OH:
                OH.write(fasta)

        return fasta

    def get_bed(self, path=None, prefix=""):
        if not prefix.endswith("_"):
            prefix += "_"

        bed = f'track description="ref:{self.refGenome}"\n'
        for i in self.oligoData.index:
            oligo = self.oligoData.loc[i, :]
            chromStart, chromEnd, sequence = oligo[:3]
            bed += f"{self.chrom}\t{chromStart}\t{chromEnd}\t"
            bed += f"{prefix}oligo_{i}\n"

        if type(None) != type(path):
            assert os.path.isdir(os.path.dirname(path))
            with open(path, "w+") as OH:
                OH.write(bed)

        return bed

    def _plot_region(self, outputDir, region):
        fig = plt.figure()

        chrom, start, stop = region
        plt.plot([start, stop], [0, 0], "k", linewidth=4.0, label="Genome")
        plt.plot(
            [start + (stop - start) / 2.0, start + (stop - start) / 2.0],
            [-1, 1],
            "r--",
            label="Window center",
        )

        plt.plot(
            [self.chromStart, self.chromEnd], [0, 0], "c-", linewidth=4.0, label="Probe"
        )
        plt.plot(
            [self.chromStart + self.size / 2.0, self.chromStart + self.size / 2.0],
            [-1, 1],
            "c--",
            label="Probe center",
        )

        plt.gca().axes.get_yaxis().set_visible(False)
        plt.suptitle("%s:%d-%.0f" % (chrom, start, stop))
        plt.xlabel("genomic coordinate [nt]")
        plt.legend(fontsize="small", loc="best")

        plt.savefig(
            os.path.join(outputDir, "window.png"), format="png", bbox_inches="tight"
        )
        plt.close(fig)

    def _plot_oligo(self, outputDir):
        fig = plt.figure(figsize=(20, 5))
        (genome_handle,) = plt.plot(
            [self.chromStart, self.chromEnd], [0, 0], "k", linewidth=4.0, label="Genome"
        )

        for i in self.oligoData.index:
            oligo = self.oligoData.loc[i, :]
            oligo_midpoint = (oligo["chromStart"] + oligo["chromEnd"]) / 2.0
            (oligo_handle,) = plt.plot(
                [oligo["chromStart"], oligo["chromEnd"]],
                [0, 0],
                "c",
                linewidth=2.0,
                label="Oligo",
            )
            (oligoCenter_handle,) = plt.plot(
                [oligo_midpoint, oligo_midpoint],
                [-0.1, 0.1],
                "c:",
                label="Oligo center",
            )

        plt.gca().axes.get_yaxis().set_visible(False)
        plt.ylim((-0.5, 0.5))

        plt.gca().axes.get_xaxis().set_ticks(
            list(
                range(
                    self.chromStart,
                    self.chromEnd,
                    max(1, int((self.chromEnd - self.chromStart) / 5.0)),
                )
            )
        )

        plt.legend(
            handles=[genome_handle, oligo_handle, oligoCenter_handle],
            fontsize="small",
            loc="best",
        )
        plt.suptitle((f"{self.chrom}:{self.chromStart}-{self.chromEnd}"))
        plt.xlabel("genomic coordinate [nt]")

        plt.savefig(
            os.path.join(outputDir, "probe.png"), format="png", bbox_inches="tight"
        )
        plt.close(fig)

    def _plot_oligo_distr(self, outputDir):
        fig = plt.figure()

        plt.plot(
            [0, self.oligoData.shape[0] - 1],
            [self.chromStart, self.chromEnd],
            "k-",
            label="Homogeneous distribution",
        )
        plt.plot(
            list(range(self.oligoData.shape[0])),
            self.oligoData["chromStart"].values,
            "r.",
            label="Oligo",
        )

        plt.suptitle(f"{self.chrom}:{self.chromStart}-{self.chromEnd}")
        plt.xlabel("oligo number")
        plt.ylabel("genomic coordinate [nt]")
        plt.legend(fontsize="small", loc="best")

        plt.savefig(os.path.join(outputDir, "oligo.png"), format="png")
        plt.close(fig)

    def _plot_oligo_distance(self, outputDir):
        fig = plt.figure()

        if self.oligoData.shape[0] > 1:
            startPositions = self.oligoData.iloc[1:, 0].values
            endPositions = self.oligoData.iloc[:-1, 1].values
            diffs = startPositions - endPositions
            plt.hist(diffs, density=1, facecolor="green", alpha=0.5)
            density = stats.calc_density(diffs, alpha=0.5)
            plt.plot(
                density["x"].tolist(),
                density["y"].tolist(),
                "b--",
                label="Density distribution",
            )

            plt.suptitle(f"{self.chrom}:{self.chromStart}-{self.chromEnd}")
            plt.legend(fontsize="small", loc="best")

        plt.xlabel("Distance between consecutive oligos [nt]")
        plt.ylabel("Density")

        plt.savefig(os.path.join(outputDir, "distance.png"), format="png")
        plt.close(fig)

    def plot(self, outputDir, region):
        assert os.path.isdir(outputDir), f'folder not found: "{outputDir}"'
        self._plot_region(outputDir, region)
        self._plot_oligo(outputDir)
        self._plot_oligo_distr(outputDir)
        self._plot_oligo_distance(outputDir)


def describe_candidate(candidate, queried_region):
    return candidate.describe(queried_region)


class ProbeFeatureTable(object):
    FEATURE_SORT = {
        "centrality": {"ascending": False},
        "size": {"ascending": True},
        "homogeneity": {"ascending": False},
    }

    def __init__(self, candidateList, queried_region, verbose=False, threads=1):
        super(ProbeFeatureTable, self).__init__()
        assert 0 < len(candidateList)

        self.data = []
        if threads != 1:
            verbose = 1 if verbose else 0
            self.data = Parallel(n_jobs=threads, backend="threading", verbose=verbose)(
                delayed(describe_candidate)(candidate, queried_region)
                for candidate in candidateList
            )
        else:
            candidateList = track(candidateList) if verbose else candidateList
            for candidate in candidateList:
                self.data.append(candidate.describe(queried_region))
        self.data = pd.concat(self.data)
        self.data.index = range(self.data.shape[0])

        self.discarded = None

    def reset(self):
        self.data = pd.concat([self.discarded, self.data])
        self.discarded = None
        self.data.sort_index(inplace=True)

    def keep(self, condition, cumulative=False):
        if not cumulative:
            self.reset()
        self.discarded = pd.concat(
            [self.discarded, self.data.loc[np.logical_not(condition), :]]
        )
        self.data = self.data.loc[condition, :]

    def filter(self, feature, thr, cumulative=False):
        assert (
            feature in self.FEATURE_SORT.keys()
        ), f'fetature "{feature}" not recognized.'

        if not cumulative:
            self.reset()

        self.rank(feature)

        best_feature = self.data[feature].values[0]
        feature_delta = best_feature * thr
        feature_range = (best_feature - feature_delta, best_feature + feature_delta)

        discardCondition = [
            self.data[feature] < feature_range[0],
            self.data[feature] > feature_range[1],
        ]

        discardCondition = np.logical_or(*discardCondition)
        self.discarded = pd.concat([self.discarded, self.data.loc[discardCondition, :]])
        self.data = self.data.loc[np.logical_not(discardCondition), :]

        return (feature_range, feature)

    def rank(self, feature):
        self.data = self.data.sort_values(
            feature, ascending=self.FEATURE_SORT[feature]["ascending"]
        )


class GenomicWindow(object):
    def __init__(self, chrom, start, size):
        super(GenomicWindow, self).__init__()
        self.chrom = chrom
        self.chromStart = start
        self.chromEnd = start + size
        self.midpoint = (self.chromStart + self.chromEnd) / 2
        self.size = size
        self.probe = None

    def __str__(self):
        s = f"[GenomicWindow]{self.chrom}:{self.chromStart}-{self.chromEnd}\n"
        s += f" Probe: {self.probe}"
        return s

    def asRegion(self):
        return (self.chrom, self.chromStart, self.chromEnd)

    def has_probe(self):
        return self.probe is not None

    def shift(self, n):
        return GenomicWindow(self.chrom, self.chromStart + n, self.size)

    def __repr__(self):
        return f"{self.chrom}:{self.chromStart}-{self.chromEnd}"


class GenomicWindowList(object):
    """Both a list genomic window and associated probe set."""

    data: List = []

    def __init__(self, windows=None):
        super(GenomicWindowList, self).__init__()
        if type(None) != type(windows):
            self.data = windows

    def __getitem__(self, i):
        return self.data[i]

    def __iter__(self):
        yield from self.data

    def __len__(self):
        return len(self.data)

    def add(self, chrom, start, size):
        self.data.append(GenomicWindow(chrom, start, size))

    def count_probes(self):
        return sum(w.has_probe() for w in self)

    def shift(self, n):
        return GenomicWindowList([window.shift(n) for window in self.data])

    def sort(self):
        midpointList = np.array([w.midpoint for w in self.data])
        self.data = [self.data[i] for i in np.argsort(midpointList)]

    def calc_probe_size_and_homogeneity(self):
        if self.count_probes() < 3:
            return np.nan

        probeData = pd.concat(
            [
                w.probe.asDataFrame((w.chromStart, w.chromEnd))
                for w in self
                if w.probe is not None
            ]
        )
        size_std = probeData["size"].values.std()
        probe_homogeneity = np.std(
            probeData["chromStart"].values[1:] - probeData["chromEnd"].values[:-1]
        )
        return 2 / (size_std + probe_homogeneity)

    def asDataFrame(self):
        starts = []
        ends = []
        sizes = []
        for window in self:
            starts.append(window.chromStart)
            ends.append(window.chromEnd)
            sizes.append(window.size)
        return pd.DataFrame.from_dict(
            {"chromStart": starts, "chromEnd": ends, "size": sizes}
        )

    def _plot_probe_set(self, outputDir, region):
        fig = plt.figure(figsize=(20, 5))

        for wi in range(len(self)):
            window = self[wi]
            plt.axvline(window.chromStart, ymin=-1, ymax=1, color="k")
            if not window.has_probe():
                plt.gca().add_patch(
                    patches.Rectangle(
                        (window.chromStart, -1), window.size, 2, color="r"
                    )
                )
            plt.gca().text(window.midpoint, 0.5, wi + 1)
        plt.axvline(self[0].chromStart, ymin=-1, ymax=1, color="k", linestyle=":")
        plt.axvline(self[-1].chromEnd, ymin=-1, ymax=1, color="k", linestyle=":")

        plt.gca().add_patch(
            patches.Rectangle(
                (region[1], -1), self[0].chromStart - region[1], 2, color="k"
            )
        )
        plt.gca().add_patch(
            patches.Rectangle(
                (self[-1].chromEnd, -1), region[2] - self[-1].chromEnd, 2, color="k"
            )
        )

        (genome,) = plt.plot(
            [region[1], region[2]], [0, 0], "k", linewidth=4.0, label="Genome"
        )

        for probe in [w.probe for w in self if w.probe is not None]:
            (oligo,) = plt.plot(
                [probe.chromStart, probe.chromEnd],
                [0, 0],
                "c",
                linewidth=2.0,
                label="Probe",
            )
            (oligo_center,) = plt.plot(
                [probe.midpoint, probe.midpoint],
                [-0.1, 0.1],
                "c:",
                label="Probe center",
                linewidth=2.0,
            )

        plt.gca().axes.get_yaxis().set_visible(False)
        plt.ylim((-1, 1))
        msg = "Empty windows are reported in red."
        plt.suptitle(
            "".join(
                [
                    f"Region: {region[0]}:{region[1]}-{region[2]}",
                    " & Probe set: ",
                    f"{self[0].chrom}:{self[0].chromStart}-{self[-1].chromEnd}\n",
                    msg,
                ]
            )
        )
        plt.xlabel("genomic coordinate [nt]")

        plt.savefig(
            os.path.join(outputDir, "windows.png"), format="png", bbox_inches="tight"
        )
        plt.close(fig)

    def _plot_probe_distr(self, outputDir):
        fig = plt.figure()

        probes = [w.probe for w in self if w.probe is not None]

        if len(probes) > 1:
            plt.plot(
                [0, len(probes) - 1],
                [probes[0].midpoint, probes[-1].midpoint],
                "k-",
                label="Homogeneous distribution",
            )
            plt.plot(
                list(range(len(probes))),
                [p.midpoint for p in probes],
                "r.",
                label="Probe",
            )

            plt.suptitle(
                f"{probes[0].chrom}:" + f"{probes[0].chromStart}{probes[-1].chromEnd}"
            )
            plt.xlim((-1, len(probes)))
            plt.legend(fontsize="small", loc="best")

        plt.xlabel("probe number")
        plt.ylabel("genomic coordinate [nt]")

        plt.savefig(
            os.path.join(outputDir, "distr.png"), format="png", bbox_inches="tight"
        )
        plt.close(fig)

    def _plot_probe_distance(self, outputDir, region):
        fig = plt.figure()

        probes = [w.probe for w in self if w.probe is not None]

        if len(probes) > 1:
            starts = np.array([p.chromStart for p in probes][1:])
            ends = np.array([p.chromEnd for p in probes][:-1])
            diffs = starts - ends
            plt.hist(diffs, density=1, facecolor="green", alpha=0.5)

            density = stats.calc_density(diffs, alpha=0.5)
            plt.plot(
                density["x"].tolist(),
                density["y"].tolist(),
                "b--",
                label="Density distribution",
            )

            plt.suptitle(
                f"{probes[0].chrom}:" + f"{probes[0].chromStart}{probes[-1].chromEnd}"
            )
            plt.legend(fontsize="small", loc="best")

        plt.xlabel("Distance between consecutive probes [nt]")
        plt.ylabel("Density")

        plt.savefig(
            os.path.join(outputDir, "distance.png"), format="png", bbox_inches="tight"
        )
        plt.close(fig)

    def export(self, path, region):
        probe_counter = 0
        fasta = ""
        bed = ""
        for window in self:
            if window.has_probe():
                probe_dirName = f"probe_{probe_counter}"
                probe_dirPath = os.path.join(path, probe_dirName)
                probe = window.probe

                assert not os.path.isfile(probe_dirPath)
                assert not os.path.isdir(probe_dirPath)
                os.mkdir(probe_dirPath)

                probe.describe(
                    window.asRegion(),
                    os.path.join(probe_dirPath, f"probe_{probe_counter}.config"),
                )
                fasta += probe.get_fasta(
                    os.path.join(probe_dirPath, f"probe_{probe_counter}.fasta"),
                    f"probe{probe_counter}",
                )
                probe_bed = probe.get_bed(
                    os.path.join(probe_dirPath, f"probe_{probe_counter}.bed"),
                    f"probe{probe_counter}",
                )
                bed += "\n".join(probe_bed.split("\n")[1:])
                probe.plot(probe_dirPath, window.asRegion())

                probe_counter += 1

        self._plot_probe_set(path, region)
        self._plot_probe_distr(path)
        self._plot_probe_distance(path, region)

        return (fasta, bed)



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import numpy as np  # type: ignore
from scipy import stats as spStats  # type: ignore


def calc_density(data, **kwargs):
    """Calculate the Gaussian KDE of the provided data series."""
    sigma = 0.2 if "sigma" not in list(kwargs.keys()) else kwargs["sigma"]
    nbins = 1000 if "nbins" not in list(kwargs.keys()) else kwargs["nbins"]

    # If only one nucleus was found
    if len(data) == 1:
        f = eval(f"lambda x: 1 if x == {data[0]} else 0")
        f = np.vectorize(f)
        return {"x": np.array([data[0]]), "y": np.array([1]), "f": f}

    data = data[np.logical_not(np.isnan(data))]
    density = spStats.gaussian_kde(data)
    density.covariance_factor = lambda: sigma
    density._compute_covariance()
    return {
        "x": np.linspace(min(data), max(data), nbins),
        "y": density(np.linspace(min(data), max(data), nbins)),
        "f": density,
    }



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import bottle as bot  # type: ignore
from typing import Dict


class Routes:
    """Routes class."""

    # Empty routes dictionary
    data: Dict = {}

    def __init__(self):
        """
        Create new routes here using the add_route method.
        """

        # Static files ---------------------------------------------------------

        route = "/<dname:re:(css|js|fonts|images|documents)>/<path>"
        self.add_route("static_file", "route", route)

        route = "/documents/<path>/mimetype/<mt1>/<mt2>"
        self.add_route("static_raw_file_download", "route", route)

        # Pages ----------------------------------------------------------------

        self.add_route("home", "route", "/")
        self.add_route("home", "view", "home.tpl.html")

        # Errors ---------------------------------------------------------------

        self.add_route("error404", "error", 404)
        self.add_route("error500", "error", 500)

        return

    def add_route(self, k, t, v):
        """Create a new route or add data to an existing one.

        Args:
                k (string): route name, identical to the route function.
                t (string): the data type (route|view|get|post|error).
                v (string|int): the data value.
        """

        # Create a new route if it does not exist
        if k not in self.data.keys():
            self.data[k] = {}

        # Update existing route
        self.data[k][t] = v

        # Output
        return

    # Pages --------------------------------------------------------------------

    def home(routes, self):
        """Default home-page.

        Args:
                self (App): ProbeDesigner.App instance.
        """
        d = self.vd.copy()
        d["title"] = "Default Home-page"
        d["description"] = "Nothing to say..."
        return d

    # Static files -------------------------------------------------------------

    def static_file(routes, self, dname, path):
        """Access static files.

        Args:
                self (App): ProbeDesigner.App instance.
                dname (string): file type.
                path (string): file name.
        """
        return bot.static_file(path, "%s/%s/" % (self.local_path, dname))

    def static_raw_file_download(routes, self, path, mt1, mt2):
        """Download raw static files.

        Args:
                self (App): ProbeDesigner.App instance.
                path (string): file name.
                mt1 (string): first part of the mimetype.
                mt2 (string): second part of the mimetype.
        """
        ipath = "%s/documents/" % self.local_path
        mt = "%s/%s" % (mt1, mt2)
        return bot.static_file(path, ipath, mimetype=mt)

    # Error --------------------------------------------------------------------

    def error404(routes, self, error):
        """Error 404.

        Args:
                self (App): ProbeDesigner.App instance.
                error: error data.
        """
        return "ERROR 404: nothing here, sorry :("

    def error500(routes, self, error):
        """Error 500.

        Args:
                self (App): ProbeDesigner.App instance.
                error: error data.
        """
        return "ERROR 500: nothing here, sorry :("



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

from importlib.metadata import version

try:
    __version__ = version(__name__)
except Exception as e:
    raise e

from ifpd import bioext, exception, query, stats
from ifpd import sections

__all__ = ["__version__", "bioext", "exception", "query", "sections", "stats"]



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import logging
import sys
from typing import Callable


def enable_rich_assert(fun: Callable) -> Callable:
    def wrapper(*args, **kwargs):
        try:
            return fun(*args, **kwargs)
        except AssertionError as e:
            logging.exception(e)
            sys.exit()

    return wrapper



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import configparser
import datetime
import logging
import os
import subprocess as sp
import threading
import time


class Enquirer(threading.Thread):
    """Database enquirer."""

    def __init__(self, queue):
        """Instance method.

        Args:
                queue (Queue): Queue object which will contain the queries.
        """

        # Initializer
        super(Enquirer, self).__init__()

        # Set class vars
        self.queue = queue

        # Close
        return

    def readQueryConfig(self, queryDir):
        config = configparser.ConfigParser()
        with open(f"{queryDir}.config", "r") as IH:
            config.read_string("".join(IH.readlines()))
        return config

    def writeQueryConfig(self, queryDir, config):
        with open(f"{queryDir}.config", "w+") as OH:
            config.write(OH)

    def __get_outdir_id(self, cmd):
        return 5

    def run(self):
        """Run one query from the queue."""

        while True:
            if not self.queue.empty():
                cmd = self.queue.get()

                # If the queue released a task
                if type(None) != type(cmd):
                    outdir_id = self.__get_outdir_id(cmd)

                    query_id = os.path.basename(cmd[outdir_id])
                    with open(f"{cmd[outdir_id]}.error.log", "w+") as EH:
                        logging.debug(f'Running query "{query_id}"')
                        timestamp = time.time()
                        isotimestamp = datetime.datetime.fromtimestamp(
                            timestamp
                        ).isoformat()
                        config = self.readQueryConfig(cmd[outdir_id])
                        config["GENERAL"]["status"] = "running"
                        config["WHEN"]["start_time"] = f"{timestamp}"
                        config["WHEN"]["start_isotime"] = isotimestamp
                        self.writeQueryConfig(cmd[outdir_id], config)

                        sp.call(cmd, stderr=EH)
                        timestamp = time.time()
                        isotimestamp = datetime.datetime.fromtimestamp(
                            timestamp
                        ).isoformat()
                        logging.debug(f'Finished query "{query_id}"')
                        config = self.readQueryConfig(cmd[outdir_id])
                        config["GENERAL"]["status"] = "done"
                        config["WHEN"]["done_time"] = f"{timestamp}"
                        config["WHEN"]["done_isotime"] = isotimestamp
                        self.writeQueryConfig(cmd[outdir_id], config)
                        cmd = self.queue.task_done(cmd)
        return



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import argparse
import configparser
from ifpd import bioext
from ifpd.scripts import arguments as ap  # type: ignore
from ifpd.exception import enable_rich_assert
import logging
import numpy as np  # type: ignore
import os
import pandas as pd  # type: ignore
from rich.logging import RichHandler  # type: ignore
import shutil

logging.basicConfig(
    level=logging.INFO,
    format="%(message)s",
    handlers=[RichHandler(markup=True, rich_tracebacks=True)],
)


def init_parser(subparsers: argparse._SubParsersAction) -> argparse.ArgumentParser:
    parser = subparsers.add_parser(
        __name__.split(".")[-1],
        description="""
Builds a database of complementary oligodeoxyribonucleotides (ODNs) in a format
compatible with ifpd.

DISCLAIMER: this script does NOT generate a new database. It only re-formats an
existing database to a format compatible with ifpd.

The script takes a tabulation-separated values (tsv) file with four (4) columns:
chromosome (or feature), chromStart, chromEnd, sequence. The first three (3) column are
identical to the columns of a BED3 file.

NOTE: only non-arbitrary nucleotides in standard IUPAC format are allowed in the
ODNs sequence. https://www.bioinformatics.org/sms/iupac.html

After the database is created, we advise running the ifpd dbchk script to
test its integrity.

Details on the BED3 format are available on the UCSC website:
https://genome.ucsc.edu/FAQ/FAQformat.html#format1
""",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        help="Builds a database of complementary ODNs compatible with FISH-ProDe.",
    )

    parser.add_argument(
        "input",
        metavar="input",
        type=str,
        help="""Path to input file. Four (4) tabulation-separated columns are expected:
        chromosome, chromStart, chromEnd, and sequence.""",
    )
    parser.add_argument("dbName", metavar="dbName", type=str, help="Database name.")

    parser.add_argument(
        "--output",
        metavar="dirPath",
        type=str,
        default=".",
        help="""Path to directory where to build the database. Created if missing.
            Defaults a subfolder of current directory with database name.""",
    )
    parser.add_argument(
        "--refGenome",
        metavar="refGen",
        type=str,
        default="",
        help="Reference genome. Only stored in the config file.",
    )
    parser = ap.add_version_option(parser)

    advanced = parser.add_argument_group("advanced arguments")
    advanced.add_argument(
        "--increment-chrom-end",
        action="store_const",
        dest="incrementChromEnd",
        const=True,
        default=False,
        help="""Use this option if your input bed file is not in UCSC bed format,
            and the last position (chromEnd) is actually included. This forces
            a unit increase of that position to convert to UCSC bed format.""",
    )
    advanced.add_argument(
        "--custom-config",
        metavar="config",
        type=str,
        nargs="*",
        help="""Space separated "key:value" pairs to store in the
        database .config file. No quotes needed.""",
    )
    advanced.add_argument(
        "-f",
        action="store_const",
        dest="forceRun",
        const=True,
        default=False,
        help="""Force overwriting of the database if already run.
            !!!This is potentially dangerous!!!""",
    )

    parser.set_defaults(parse=parse_arguments, run=run)

    return parser


@enable_rich_assert
def parse_arguments(args: argparse.Namespace) -> argparse.Namespace:
    if args.output == ".":
        args.output = args.dbName

    assert os.path.isfile(args.input), f'input file not found: "{args.input}"'
    assert not os.path.isfile(
        args.output
    ), f'expected output path to a folder, file found: "{args.output}"'

    if args.forceRun:
        if os.path.isdir(args.output):
            logging.warning("'-f' option used. Removing previously generated database.")
            shutil.rmtree(args.output)
    else:
        assert not os.path.isdir(
            args.output
        ), f'output folder already exists: "{args.output}"'
    return args


def sort_oligos(args, chromList):
    logging.info("Sort oligos.")
    has_overlaps = False
    oligoLengthRange = [np.inf, -np.inf]
    oligoMinDist = np.inf
    for chrom in chromList:
        chromDF = pd.read_csv(os.path.join(args.output, chrom), "\t", header=None)
        chromDF.columns = bioext.UCSCbed.FIELD_NAMES[1 : (chromDF.shape[1] + 1)]

        chromDF = chromDF.sort_values("chromStart")
        chromDF.to_csv(
            os.path.join(args.output, chrom), "\t", header=False, index=False
        )

        startPositions = np.array(chromDF["chromStart"].iloc[1:].tolist())
        endPositions = np.array(chromDF["chromEnd"].iloc[:-1].tolist()) - 1
        if any(startPositions <= endPositions):
            has_overlaps = True

        oligoLengthList = np.unique(chromDF["chromEnd"] - chromDF["chromStart"])
        oligoLengthRange[0] = min(oligoLengthRange[0], oligoLengthList.min())
        oligoLengthRange[1] = max(oligoLengthRange[1], oligoLengthList.max())

        assert len(startPositions) == len(endPositions)
        if len(startPositions) == 0:
            continue
        oligoMinDist = min(oligoMinDist, min(startPositions - endPositions - 1))
    return oligoMinDist, oligoLengthRange, has_overlaps


def mk_config(args, oligoMinDist, oligoLengthRange, has_overlaps):
    if np.inf == oligoMinDist:
        oligoMinDist = 0
    logging.info("Write config file.")
    config = configparser.ConfigParser()
    config["DATABASE"] = {"name": args.dbName, "refGenome": args.refGenome}
    config["OLIGOS"] = {
        "min_dist": oligoMinDist,
        "min_length": oligoLengthRange[0],
        "max_length": oligoLengthRange[1],
        "overlaps": str(has_overlaps),
    }
    config["SOURCE"] = {
        "bed": args.input,
        "outDirectory": args.output,
    }
    if args.custom_config is not None:
        config["CUSTOM"] = {}
        for x in args.custom_config:
            k, v = x.split(":")[:2]
            config["CUSTOM"][k] = v

    with open(os.path.join(args.output, ".config"), "w+") as OH:
        config.write(OH)


@enable_rich_assert
def run(args: argparse.Namespace) -> None:
    os.mkdir(args.output)

    logging.info("Set up input buffer.")
    oligoGenerator = bioext.UCSCbed(
        args.input, incrementChromEnd=args.incrementChromEnd, bufferize=True
    ).buffer(False, False)

    logging.info("Parse and write database.")
    chromList = set()
    for line in oligoGenerator:
        if len(line.strip()) == 0:
            continue
        line = line.strip().split("\t")
        chrom = line.pop(0)
        line = "\t".join(line) + "\n"
        chromList.add(chrom)
        with open(os.path.join(args.output, chrom), "a+") as OH:
            OH.write(line)

    mk_config(args, *sort_oligos(args, chromList))
    logging.info(f'Created database "{args.dbName}".')

    logging.info("Done. :thumbs_up: :smiley:")



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

from ifpd.sections.probe_design.app import App
from ifpd.sections.probe_design.enquirer import Enquirer
from ifpd.sections.probe_design.query import Query
from ifpd.sections.probe_design.queue import Queue
from ifpd.sections.probe_design.routes import Routes

__all__ = ["App", "Enquirer", "Query", "Queue", "Routes"]



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import queue as q
from typing import List


class Queue(q.Queue):
    """Database query Queue.

    Args:
            MAX_CURR (int): maximum number of simultaneously released tasks.
            doing (list): list of currently released tasks (i.e., running).
            done (list): list of completed tasks.
    """

    MAX_CURR = 1
    doing: List = []
    done: List = []

    def __init__(self, MAX_CURR=None, **kwargs):
        """Instance method.

        Args:
                n_doing (int): max number of simultaneously running tasks.
        """
        if type(None) != type(MAX_CURR):
            self.MAX_CURR = MAX_CURR
        super(Queue, self).__init__()
        return

    def get(self, **kwargs):
        """Extend original get method by setting up doing.
        Also forces only one element to be running at a time.
        """

        # Stop if already running
        if self.MAX_CURR < len(self.doing):
            return

        # Call original method
        released = super(Queue, self).get(**kwargs)

        # Append released element to released list
        self.doing.append(released)

        # Output released element
        return released

    def task_done(self, done, **kwargs):
        """Extend original task_done method by adding doing and done features."""

        # If unknown task, kill execution
        if done not in self.doing:
            return

        # Set the task as completed
        self.done.append(self.doing.pop(self.doing.index(done)))

        # Call original method
        super(Queue, self).task_done(**kwargs)

        # Stop
        return



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import argparse
from ifpd import const, query
from ifpd.scripts import arguments as ap  # type: ignore
from ifpd.exception import enable_rich_assert
import logging
import numpy as np  # type: ignore
import os
from rich.logging import RichHandler  # type: ignore
from rich.progress import track  # type: ignore
import shutil

logging.basicConfig(
    level=logging.INFO,
    format="%(message)s",
    handlers=[RichHandler(markup=True, rich_tracebacks=True)],
)


def init_parser(subparsers: argparse._SubParsersAction) -> argparse.ArgumentParser:
    parser = subparsers.add_parser(
        __name__.split(".")[-1],
        description="""
Design a FISH probe in a genomic region of interest using a database of
oligonucleotides. Concisely, the script does the following:
- Identify all probe candidates
- Calculate centrality, size, and homogeneity for each candidate
- Identify the candidate with the best first feature (featOrder), i.e.,
  max(centrality), min(size), or max(homogeneity).
- Build a range around the best first feature value (filterThr) and filter out
  all the candidates that do not fall in it.
- Rank the remaining candidates based on the second feature (featOrder), i.e.,
  decreasing for centrality, increasing for size or homogeneity.
- Return the top N candidates (maxProbes), with plots, tables, fasta and bed.
""",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        help="Design a FISH probe in a genomic region of interest.",
    )

    parser.add_argument(
        "database", metavar="database", type=str, help="Path to database folder."
    )
    parser.add_argument(
        "chrom",
        type=str,
        help="Database feature to query for a probe.",
    )
    parser.add_argument(
        "outdir",
        metavar="outDir",
        type=str,
        help="Path to query output directory. Stops if it exists already.",
    )

    parser.add_argument(
        "--region",
        metavar=("chromStart", "chromEnd"),
        type=int,
        nargs=2,
        default=(0, np.inf),
        help="""Start and end locations (space-separated) of the region of interest.
        When a region is not provided (or start/end coincide),
        the whole feature is queried.""",
    )
    parser.add_argument(
        "--n-oligo",
        metavar="nOligo",
        type=int,
        default=48,
        help="""Number of oligos per probe. If not enough oligos are found, the largest
        probe (with greater number of oligos) is designed by default. Default: 48""",
    )
    parser.add_argument(
        "--max-probes",
        metavar="nProbes",
        type=int,
        default=-1,
        help="""Maximum number of probe candidates to output.
            Set to -1 to retrieve all candidates. Default: -1""",
    )

    parser = ap.add_version_option(parser)

    advanced = parser.add_argument_group("advanced arguments")
    advanced.add_argument(
        "--order",
        metavar="feature",
        type=str,
        default=const.featureList,
        nargs="+",
        help="""Space-separated features, used as explained in script description.
        The available features are: 'centrality', 'size', and 'homogeneity'. At least 2
        features must be listed. Default: "size homogeneity centrality".""",
    )
    advanced.add_argument(
        "--filter-thr",
        metavar="filterThr",
        type=float,
        default=0.1,
        help="""Threshold of first feature filter, used to identify
        a range around the best value (percentage range around it). Accepts values
        from 0 to 1. Default: 0.1""",
    )
    advanced.add_argument(
        "--min-d",
        metavar="minD",
        type=int,
        default=10,
        help="*DEPRECATED* Minimum distance between consecutive oligos. Default: 10",
    )
    advanced.add_argument(
        "--exact-n-oligo",
        action="store_const",
        dest="exact_n_oligo",
        const=True,
        default=False,
        help="""Stop if not enough oligos are found,
        instead of designing the largest probe.""",
    )
    advanced.add_argument(
        "-f",
        action="store_const",
        dest="forceRun",
        const=True,
        default=False,
        help="""Force overwriting of the query if already run.
            !!!This is potentially dangerous!!!""",
    )

    parser.set_defaults(parse=parse_arguments, run=run)

    return parser


def assert_region(args):
    if args.region is not None:
        if args.region[0] == args.region[1]:
            args.region = None
            return
        assert (
            args.region[0] >= 0
        ), f"start location cannot be negative [{args.region[0]}]."
        assert (
            args.region[1] >= 0
        ), f"end location cannot be negative [{args.region[1]}]."
        assert (
            args.region[1] > args.region[0]
        ), f"end location must be greater than start location [{args.region}]."


@enable_rich_assert
def parse_arguments(args: argparse.Namespace) -> argparse.Namespace:
    assert not os.path.isfile(
        args.outdir
    ), f"output folder expected, file found: {args.outdir}"
    if args.forceRun:
        if os.path.isdir(args.outdir):
            shutil.rmtree(args.outdir)
            logging.warning("Overwriting previously run query.")
    else:
        assert not os.path.isdir(
            args.outdir
        ), f"output folder already exists: {args.outdir}"
    assert_region(args)

    assert 2 <= len(
        args.order
    ), f"at least 2 features needed, only {len(args.order)} found."

    for o in args.order:
        assert (
            o in const.featureList
        ), f'unrecognized feature "{o}". Should be one of {const.featureList}.'

    assert (
        args.filter_thr >= 0 and args.filter_thr <= 1
    ), "first filter threshold must be a fraction: {args.filter_thr}"

    assert (
        args.min_d >= 0
    ), f"negative minimum distance between consecutive oligos: {args.min_d}"

    assert args.n_oligo >= 1, f"a probe must have oligos: {args.n_oligo}"
    if args.max_probes == -1:
        args.max_probes = np.inf
    assert args.max_probes >= 0, f"at least 1 probe in output: {args.max_probes}"

    return args


@enable_rich_assert
def run(args: argparse.Namespace) -> None:
    os.mkdir(args.outdir)
    ap.add_log_file_handler(os.path.join(args.outdir, "log"))

    logging.info("Read database.")
    oligoDB = query.OligoDatabase(args.database)

    assert (
        not oligoDB.has_overlaps()
    ), "databases with overlapping oligos are not supported yet."
    assert oligoDB.has_chromosome(
        args.chrom
    ), f'chromosome "{args.chrom}" not in the database.'

    oligoDB.read_chromosome(args.chrom)
    chromData = oligoDB.chromData[args.chrom]
    if args.region[1] == np.inf:
        args.region = (args.region[0], chromData["chromEnd"].max())
    chromStart, chromEnd = args.region
    queried_region = (args.chrom, chromStart, chromEnd)

    selectCondition = np.logical_and(
        chromData.iloc[:, 0] >= chromStart, chromData.iloc[:, 1] <= chromEnd
    )
    selectedOligos = chromData.loc[selectCondition, :]

    args = ap.check_n_oligo(args, selectCondition)

    logging.info("Build probe candidates.")
    candidateList = [
        query.OligoProbe(
            queried_region[0],
            selectedOligos.iloc[i : (i + args.n_oligo), :],
            oligoDB,
        )
        for i in track(
            range(selectedOligos.shape[0] - args.n_oligo + 1),
            description="Building",
        )
    ]

    logging.info(f"Found {len(candidateList)} probe candidates.")

    logging.info("Describing candidates...")
    probeFeatureTable = query.ProbeFeatureTable(candidateList, queried_region, True)
    feature_range, feature = probeFeatureTable.filter(args.order[0], args.filter_thr)
    feature_range = np.round(feature_range, 6)
    logging.info(
        "".join(
            [
                f"Selected {probeFeatureTable.data.shape[0]} candidates ",
                f"in the range {feature_range} of '{feature}'.",
            ]
        )
    )

    logging.info(f"Ranking based on '{args.order[1]}'.")
    probeFeatureTable.rank(args.order[1])

    logging.info("Writing description table...")
    probeFeatureTable.data.to_csv(
        os.path.join(args.outdir, "candidates.tsv"), "\t", index=False
    )

    if np.isfinite(args.max_probes):
        logging.info(f"Exporting top {args.max_probes} candidates...")
    else:
        logging.info("Exporting candidates...")
    for i in range(min(args.max_probes, probeFeatureTable.data.shape[0])):
        candidate = candidateList[probeFeatureTable.data.index[i]]
        candidatePath = os.path.join(args.outdir, f"candidate_{i}")
        os.mkdir(candidatePath)
        candidate.describe(
            queried_region, os.path.join(candidatePath, f"candidate_{i}.config")
        )
        candidate.get_fasta(os.path.join(candidatePath, f"candidate_{i}.fasta"))
        candidate.get_bed(os.path.join(candidatePath, f"candidate_{i}.bed"))
        candidate.plot(candidatePath, queried_region)

    logging.info("Done. :thumbs_up: :smiley:")



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import bottle as bot  # type: ignore


class App(bot.Bottle):
    """WebServer section App."""

    def __init__(self):
        """Initialize."""

        # Run default initialization
        super(App, self).__init__()

        # Output
        return

    def __get_proper_rout(self, data, page):
        # Load template if requested
        if "view" in data.keys():

            @self.route(data["route"])
            @bot.view(self.vpath + data["view"])
            def route(**kwargs):
                return getattr(self.route_list, page)(self, **kwargs)

        else:

            @self.route(data["route"])
            def route(**kwargs):
                return getattr(self.route_list, page)(self, **kwargs)

        return route

    def __get_reception(self, data, page):
        @self.get(data["get"])
        def route(**kwargs):
            return getattr(self.route_list, page)(self, **kwargs)

        return route

    def __post_reception(self, data, page):
        @self.post(data["post"])
        def route(**kwargs):
            return getattr(self.route_list, page)(self, **kwargs)

        return route

    def __error_page(self, data, page):
        @self.error(data["error"])
        def route(*args, **kwargs):
            return getattr(self.route_list, page)(self, *args, **kwargs)

        return route

    def route_builder(self, page, data):
        """Generate routes based on route data.

        Args:
                page (string): route name, identical to route function name.
                data (dict): rout data.
        """

        # If a proper route is requested
        if "route" in data.keys():
            route = self.__get_proper_rout(data, page)

        # Get reception
        elif "get" in data.keys():
            route = self.__get_reception(data, page)

        # Post reception
        elif "post" in data.keys():
            route = self.__post_reception(data, page)

        # Error pages
        elif "error" in data.keys():
            route = self.__error_page(data, page)

        else:
            return ()

        return route

    def build_routes(self):
        """Build routes."""

        return {
            page: self.route_builder(page, self.route_list.data[page])
            for page in self.route_list.data.keys()
        }



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import bottle as bot  # type: ignore
import configparser
import datetime
import hashlib
import ifpd as fp
from ifpd.sections import routes
from ifpd.sections.probe_design.query import Query
import logging
import os
import pandas as pd  # type: ignore
import shlex
import time
from typing import Dict
import zipfile


def zipFile(path, ziph, root=None):
    """Zips a file.
    Args:
            path (string): file path.
            ziph (Zipfile): zipfile handler.
    Returns:
            None.
    """
    assert os.path.isfile(path), "file expected."
    oldRoot = ""
    new_path = path
    if type(None) != type(root):
        oldRoot = os.path.commonpath([root, path])
        if len(oldRoot) != 0 and new_path.startswith(oldRoot):
            new_path = new_path[len(oldRoot) :]
    ziph.write(path, new_path)


def zipDir(path, ziph, mainRoot=None):
    """Zips a directory. Based on https://goo.gl/GeCPVm.
    Args:
            path (string): directory path.
            ziph (Zipfile): zipfile handler.
    Returns:
            None.
    """

    assert os.path.isdir(path), "folder expected."
    for root, dirs, files in os.walk(path):
        for file in files:
            zipFile(os.path.join(root, file), ziph, mainRoot)


class Routes(routes.Routes):
    """Routes class."""

    # Empty routes dictionary
    data: Dict = {}
    zipDirName = "zips"

    def __init__(self):
        """
        Create new routes here using the add_route method.
        """

        # Set default Routes
        super(Routes, self).__init__()

        # Static files ---------------------------------------------------------

        route = "/q/<query_id>/download/"
        self.add_route("query_download", "route", route)

        dname = ("<dname:re:(images|documents)>",)
        route = "/q/<query_id>/c/<candidate_id>/%s/<path>" % dname
        self.add_route("candidate_static_file", "route", route)
        route = "/q/<query_id>/c/<candidate_id>/documents/"
        route += "<path:re:.*>/download/"
        self.add_route("candidate_static_file_download", "route", route)
        route = "/q/<query_id>/c/<candidate_id>/download/"
        self.add_route("candidate_download", "route", route)

        dname = ("<dname:re:(images|documents)>",)
        route = "/q/<query_id>/cs/<candidate_id>/%s/<path>" % dname
        self.add_route("candidate_set_static_file", "route", route)
        route = "/q/<query_id>/cs/<candidate_id>/documents/"
        route += "<path:re:.*>/download/"
        self.add_route("candidate_set_static_file_download", "route", route)
        route = "/q/<query_id>/cs/<candidate_id>/download/"
        self.add_route("candidate_set_download", "route", route)

        dname = ("<dname:re:(images|documents)>",)
        route = "/q/<query_id>/cs/<candidate_id>/p/<probe_id>/%s/<path>" % dname
        self.add_route("candidate_set_probe_static_file", "route", route)
        route = "/q/<query_id>/cs/<candidate_id>/p/<probe_id>/documents/"
        route += "<path:re:.*>/download/"
        self.add_route("candidate_set_probe_static_file_download", "route", route)
        route = "/q/<query_id>/cs/<candidate_id>/p/<probe_id>/download/"
        self.add_route("candidate_set_probe_download", "route", route)

        # Pages ----------------------------------------------------------------

        self.add_route("home", "route", "")
        self.add_route("home", "route", "/")
        self.add_route("home", "view", "home.tpl.html")

        self.add_route("query", "route", "/q/<query_id>")
        self.add_route("query", "view", "query.tpl.html")

        uri = "/q/<query_id>/c/<candidate_id>"
        self.add_route("candidate_probe", "route", uri)
        self.add_route("candidate_probe", "view", "candidate_probe.tpl.html")

        uri = "/q/<query_id>/cs/<candidate_id>"
        self.add_route("candidate_set", "route", uri)
        self.add_route("candidate_set", "view", "candidate_set.tpl.html")

        uri = "/q/<query_id>/cs/<candidate_id>/p/<probe_id>"
        self.add_route("candidate_set_probe", "route", uri)
        self.add_route("candidate_set_probe", "view", "candidate_set_probe.tpl.html")

        # Forms ----------------------------------------------------------------

        self.add_route("single_query", "post", "/single_query")

        self.add_route("spotting_query", "post", "/spotting_query")

        self.add_route("single_queries", "post", "/single_queries")

        self.add_route("hide_alert", "post", "/hide_alert")

        # Errors ---------------------------------------------------------------

        self.add_route("error404", "error", 404)
        self.add_route("error500_with_redirect", "error", 500)

        # AJAX requests --------------------------------------------------------

        self.add_route("list_chromosomes", "get", "/listChr/<dbDir>")
        self.add_route("queueStatus", "get", "/queueStatus")

        return

    def mkZipDir(routes, self):
        zipDirPath = os.path.join(self.static_path, "query", routes.zipDirName)
        if not os.path.isdir(zipDirPath):
            os.mkdir(zipDirPath)

    def zipQuery(routes, self, query_id):
        zipDirPath = os.path.join(self.static_path, "query", routes.zipDirName)
        zipPath = os.path.join(zipDirPath, f"{query_id}.zip")
        zipf = zipfile.ZipFile(zipPath, "w", zipfile.ZIP_DEFLATED)
        zipDir(
            os.path.join(self.static_path, "query", query_id),
            zipf,
            os.path.join(self.static_path, "query"),
        )
        zipFile(
            os.path.join(self.static_path, "query", f"{query_id}.config"),
            zipf,
            os.path.join(self.static_path, "query"),
        )
        zipf.close()

    def zipCandidate(routes, self, query_id, candidate_id):
        zipDirPath = os.path.join(self.static_path, "query", routes.zipDirName)
        zipPath = os.path.join(zipDirPath, f"{query_id}.candidate_{candidate_id}.zip")
        zipf = zipfile.ZipFile(zipPath, "w", zipfile.ZIP_DEFLATED)
        zipDir(
            os.path.join(
                self.static_path, "query", query_id, f"candidate_{candidate_id}"
            ),
            zipf,
            os.path.join(
                self.static_path, "query", query_id, f"candidate_{candidate_id}"
            ),
        )
        zipf.close()

    def zipCandidateSet(routes, self, query_id, candidate_id):
        zipDirPath = os.path.join(self.static_path, "query", routes.zipDirName)
        zipPath = os.path.join(zipDirPath, f"{query_id}.probe_set_{candidate_id}.zip")
        zipf = zipfile.ZipFile(zipPath, "w", zipfile.ZIP_DEFLATED)
        zipDir(
            os.path.join(
                self.static_path, "query", query_id, f"probe_set_{candidate_id}"
            ),
            zipf,
            os.path.join(
                self.static_path, "query", query_id, f"probe_set_{candidate_id}"
            ),
        )
        zipf.close()

    def zipCandidateSetProbe(routes, self, query_id, candidate_id, probe_id):
        zipDirPath = os.path.join(self.static_path, "query", routes.zipDirName)
        zipPath = os.path.join(
            zipDirPath, f"{query_id}.probe_set_{candidate_id}.probe_{probe_id}.zip"
        )
        zipf = zipfile.ZipFile(zipPath, "w", zipfile.ZIP_DEFLATED)
        zipDir(
            os.path.join(
                self.static_path,
                "query",
                query_id,
                f"probe_set_{candidate_id}",
                f"probe_{probe_id}",
            ),
            zipf,
            os.path.join(
                self.static_path,
                "query",
                query_id,
                f"probe_set_{candidate_id}",
                f"probe_{probe_id}",
            ),
        )
        zipf.close()

    # Static files -------------------------------------------------------------

    def candidate_static_file(routes, self, query_id, candidate_id, dname, path):
        """Access candidate static files.

        Args:
                self (App): ProbeDesigner.App instance.
                query_id (string): query folder name.
                candidate_id (string): candidate folder name.
                dname (string): file type.
                path (string): file name.
        """
        ipath = "%s/query/%s/candidate_%s/" % (self.static_path, query_id, candidate_id)
        return bot.static_file(path, ipath)

    def candidate_static_file_download(routes, self, query_id, candidate_id, path):
        """Download candidate static files.

        Args:
                self (App): ProbeDesigner.App instance.
                query_id (string): query folder name.
                candidate_id (string): candidate folder name.
                dname (string): file type.
                path (string): file name.
        """
        ipath = "%s/query/%s/candidate_%s/" % (self.static_path, query_id, candidate_id)
        outname = "%s.%s" % (query_id, path)
        return bot.static_file(path, ipath, download=outname)

    def candidate_set_static_file(routes, self, query_id, candidate_id, dname, path):
        """Access candidate static files.

        Args:
                self (App): ProbeDesigner.App instance.
                query_id (string): query folder name.
                candidate_id (string): candidate folder name.
                dname (string): file type.
                path (string): file name.
        """
        ipath = "%s/query/%s/probe_set_%s/" % (self.static_path, query_id, candidate_id)
        return bot.static_file(path, ipath)

    def candidate_set_static_file_download(routes, self, query_id, candidate_id, path):
        """Download candidate static files.

        Args:
                self (App): ProbeDesigner.App instance.
                query_id (string): query folder name.
                candidate_id (string): candidate folder name.
                dname (string): file type.
                path (string): file name.
        """
        ipath = "%s/query/%s/probe_set_%s/" % (self.static_path, query_id, candidate_id)
        outname = "%s.%s" % (query_id, path)
        return bot.static_file(path, ipath, download=outname)

    def candidate_set_probe_static_file(
        routes, self, query_id, candidate_id, probe_id, dname, path
    ):
        """Access candidate static files.

        Args:
                self (App): ProbeDesigner.App instance.
                query_id (string): query folder name.
                candidate_id (string): candidate folder name.
                dname (string): file type.
                path (string): file name.
        """
        ipath = "%s/query/%s/probe_set_%s/probe_%s/" % (
            self.static_path,
            query_id,
            candidate_id,
            probe_id,
        )
        return bot.static_file(path, ipath)

    def candidate_set_probe_static_file_download(
        routes, self, query_id, candidate_id, probe_id, path
    ):
        """Download candidate static files.

        Args:
                self (App): ProbeDesigner.App instance.
                query_id (string): query folder name.
                candidate_id (string): candidate folder name.
                dname (string): file type.
                path (string): file name.
        """
        ipath = "%s/query/%s/probe_set_%s/probe_%s/" % (
            self.static_path,
            query_id,
            candidate_id,
            probe_id,
        )
        outname = "%s.probe_set_%s.%s" % (query_id, candidate_id, path)
        return bot.static_file(path, ipath, download=outname)

    def query_download(routes, self, query_id):
        """Download compressed query output.

        Args:
                self (App): ProbeDesigner.App instance.
                query_id (string): query folder name.
        """

        routes.mkZipDir(self)
        ipath = os.path.join(self.static_path, "query", routes.zipDirName)
        fName = f"{query_id}.zip"
        outname = f"query.{fName}"

        if not os.path.isfile(os.path.join(ipath, fName)):
            routes.zipQuery(self, query_id)

        return bot.static_file(fName, ipath, download=outname)

    def candidate_download(routes, self, query_id, candidate_id):
        """Download compressed candidate output.

        Args:
                self (App): ProbeDesigner.App instance.
                query_id (string): query folder name.
                candidate_id (string): candidate folder name.
        """

        routes.mkZipDir(self)
        ipath = os.path.join(self.static_path, "query", routes.zipDirName)
        fName = f"{query_id}.candidate_{candidate_id}.zip"
        outname = f"query.{fName}"

        if not os.path.isfile(os.path.join(ipath, fName)):
            routes.zipCandidate(self, query_id, candidate_id)

        return bot.static_file(fName, ipath, download=outname)

    def candidate_set_download(routes, self, query_id, candidate_id):
        """Download compressed candidate output.

        Args:
                self (App): ProbeDesigner.App instance.
                query_id (string): query folder name.
                candidate_id (string): candidate folder name.
        """

        routes.mkZipDir(self)
        ipath = os.path.join(self.static_path, "query", routes.zipDirName)
        fName = f"{query_id}.probe_set_{candidate_id}.zip"
        outname = f"query.{fName}"

        if not os.path.isfile(os.path.join(ipath, fName)):
            routes.zipCandidateSet(self, query_id, candidate_id)

        return bot.static_file(fName, ipath, download=outname)

    def candidate_set_probe_download(routes, self, query_id, candidate_id, probe_id):
        """Download compressed candidate output.

        Args:
                self (App): ProbeDesigner.App instance.
                query_id (string): query folder name.
                candidate_id (string): candidate folder name.
        """

        routes.mkZipDir(self)
        ipath = os.path.join(self.static_path, "query", routes.zipDirName)
        fName = f"{query_id}.probe_set_{candidate_id}.probe_{probe_id}.zip"
        outname = f"query.{fName}"

        if not os.path.isfile(os.path.join(ipath, fName)):
            routes.zipCandidateSetProbe(self, query_id, candidate_id, probe_id)

        return bot.static_file(fName, ipath, download=outname)

    # Pages --------------------------------------------------------------------

    def home(routes, self):
        """Home-page.

        Args:
                self (App): ProbeDesigner.App instance.
        """

        d = self.vd

        d["title"] = self.tprefix + "Home"
        d["custom_stylesheets"] = ["home.css", "style.css"]
        d["custom_root_stylesheets"] = []

        configPathList = [
            os.path.join(self.static_path, "db", p, ".config")
            for p in next(os.walk(os.path.join(self.static_path, "db")))[1]
        ]
        configPathList.sort()
        d["dbdata"] = []
        for configPath in configPathList:
            with open(configPath, "r") as IH:
                parser = configparser.ConfigParser()
                parser.read_string("".join(IH.readlines()))
                d["dbdata"].append(parser)
        d["dblist"] = {}
        for p in d["dbdata"]:
            dbDir = os.path.basename(p["SOURCE"]["outdirectory"])
            dbName = p["DATABASE"]["name"]
            d["dblist"][dbName] = dbDir

        return d

    def query(routes, self, query_id):
        """Query output page.

        Args:
                self (App): ProbeDesigner.App instance.
                query_id (string): query folder name.
        """

        d = self.vd

        d["title"] = "%s Query: %s" % (self.tprefix, query_id)
        d["custom_stylesheets"] = ["query.css", "style.css"]
        d["custom_root_stylesheets"] = []

        d["query"] = Query(query_id, self.qpath).data
        d["queryRoot"] = self.qpath

        if d["query"]["status"] == "done":
            if d["query"]["type"] == "single":
                fpath = os.path.join(self.qpath, query_id, "candidates.tsv")
            elif d["query"]["type"] == "spotting":
                fpath = os.path.join(self.qpath, query_id, "set_candidates.tsv")
            if not os.path.isfile(fpath):
                d["query"]["status"] = "error"
            else:
                d["query"]["candidate_table"] = pd.read_csv(fpath, "\t")

        d["queryTimeout"] = 24 * 60 * 60  # 1 day timeout

        d["admin_email"] = self.admin_email

        return d

    def candidate_probe(routes, self, query_id, candidate_id):
        """Candidate output page.

        Args:
                self (App): ProbeDesigner.App instance.
                query_id (string): query folder name.
                candidate_id (string): candidate folder name.
        """

        d = self.vd

        d["title"] = "%s Query: %s" % (self.tprefix, query_id)
        d["custom_stylesheets"] = ["query.css", "style.css"]
        d["custom_root_stylesheets"] = []

        d["query"] = Query(query_id, self.qpath).data
        d["queryRoot"] = self.qpath

        d["candidate"] = {"id": candidate_id}
        configPath = os.path.join(
            self.qpath,
            query_id,
            f"candidate_{candidate_id}",
            f"candidate_{candidate_id}.config",
        )
        with open(configPath, "r") as IH:
            config = configparser.ConfigParser()
            config.read_string("".join(IH.readlines()))
            d["candidate"].update(config["REGION"].items())
            d["candidate"].update(config["PROBE"].items())
            d["candidate"].update(config["FEATURES"].items())

        return d

    def candidate_set(routes, self, query_id, candidate_id):
        """Candidate output page.

        Args:
                self (App): ProbeDesigner.App instance.
                query_id (string): query folder name.
                candidate_id (string): candidate folder name.
        """

        d = self.vd

        d["title"] = "%s Query: %s" % (self.tprefix, query_id)
        d["custom_stylesheets"] = ["query.css", "style.css"]
        d["custom_root_stylesheets"] = []

        d["query"] = Query(query_id, self.qpath).data
        d["queryRoot"] = self.qpath
        d["query"]["candidate_table"] = pd.read_csv(
            os.path.join(self.qpath, query_id, "set_candidates.tsv"), "\t"
        )

        d["candidate"] = {"id": candidate_id}

        return d

    def candidate_set_probe(routes, self, query_id, candidate_id, probe_id):
        """Candidate output page.

        Args:
                self (App): ProbeDesigner.App instance.
                query_id (string): query folder name.
                candidate_id (string): candidate folder name.
        """

        d = self.vd

        d["title"] = "%s Query: %s" % (self.tprefix, query_id)
        d["custom_stylesheets"] = ["query.css", "style.css"]
        d["custom_root_stylesheets"] = []

        d["query"] = Query(query_id, self.qpath).data
        d["queryRoot"] = self.qpath
        d["query"]["candidate_table"] = pd.read_csv(
            os.path.join(self.qpath, query_id, "set_candidates.tsv"), "\t"
        )

        d["candidate"] = {"id": candidate_id}

        d["probe"] = {"id": probe_id}
        configPath = os.path.join(
            self.qpath,
            query_id,
            f"probe_set_{candidate_id}",
            f"probe_{probe_id}",
            f"probe_{probe_id}.config",
        )
        with open(configPath, "r") as IH:
            config = configparser.ConfigParser()
            config.read_string("".join(IH.readlines()))
            d["probe"].update(config["REGION"].items())
            d["probe"].update(config["PROBE"].items())
            d["probe"].update(config["FEATURES"].items())

        return d

    # Form reception -----------------------------------------------------------

    def hide_alert(routes, self):
        """Hide bookmark alert for a specific query.

        Args:
                self (App): ProbeDesigner.App instance.
        """

        formData = bot.request.forms

        Query(formData.query_id, self.qpath).data
        config = configparser.ConfigParser()
        configPath = os.path.join(self.qpath, f"{formData.query_id}.config")
        with open(configPath, "r") as IH:
            config.read_string("".join(IH.readlines()))
        config["GENERAL"]["hidden_bookmark_alter"] = "True"
        with open(configPath, "w+") as OH:
            config.write(OH)

        return "Done"

    def single_query(routes, self):
        """Single probe query form reception route.

        Args:
                self (App): ProbeDesigner.App instance.
        """

        formData = bot.request.forms
        queriedRegion = []
        if formData.start != formData.end:
            queriedRegion = [
                "--region",
                shlex.quote(formData.start),
                shlex.quote(formData.end),
            ]
        query_id = "%s:%s:%s:%s" % (
            formData.chromosome,
            formData.start,
            formData.end,
            time.time(),
        )
        encoder = hashlib.sha256()
        encoder.update(bytes(query_id, "utf-8"))
        query_id = encoder.hexdigest()

        dbPath = f"{self.static_path}/db/{formData.database}"
        oligoDB = fp.query.OligoDatabase(dbPath)
        min_dist = oligoDB.get_oligo_min_dist()

        cmd = [
            "ifpd",
            "query",
            "probe",
            shlex.quote(dbPath),
            shlex.quote(formData.chromosome),
            shlex.quote(f"{self.static_path}/query/{query_id}"),
            "--order",
            shlex.quote(formData.f1),
            shlex.quote(formData.f2),
            shlex.quote(formData.f3),
            "--filter-thr",
            shlex.quote(formData.f1_threshold),
            "--n-oligo",
            shlex.quote(formData.n_oligo),
            "--max-probes",
            shlex.quote(formData.max_probes),
            "--min-d",
            shlex.quote(f"{min_dist}"),
        ]
        if len(queriedRegion) != 0:
            cmd.extend(queriedRegion)
        logging.info(" ".join(cmd))

        config = configparser.ConfigParser()
        timestamp = time.time()
        config["GENERAL"] = {
            "name": formData.name,
            "description": formData.description,
            "type": "single",
            "cmd": " ".join(cmd),
            "status": "queued",
        }
        config["WHEN"] = {
            "time": timestamp,
            "isotime": datetime.datetime.fromtimestamp(timestamp).isoformat(),
        }
        config["WHERE"] = {"db": formData.database, "region": queriedRegion}
        config["WHAT"] = {
            "n_oligo": formData.n_oligo,
            "threshold": formData.f1_threshold,
            "max_probes": formData.max_probes,
        }
        config["HOW"] = {"f1": formData.f1, "f2": formData.f2, "f3": formData.f3}
        configPath = os.path.join(self.static_path, "query", f"{query_id}.config")
        with open(configPath, "w+") as OH:
            config.write(OH)

        self.queue.put(cmd)

        bot.response.status = 303
        bot.response.set_header(
            "Location", f"{self.root_uri}{self.app_uri}q/{query_id}"
        )

        return "Query received."

    def spotting_query(routes, self):
        """Multi probe query form reception route.

        Args:
                self (App): ProbeDesigner.App instance.
        """

        formData = bot.request.forms
        queriedRegion = []
        if formData.start != formData.end:
            queriedRegion = [
                "--region",
                shlex.quote(formData.start),
                shlex.quote(formData.end),
            ]
        query_id = "%s:%s:%s:%s" % (
            formData.multi_chromosome,
            formData.multi_start,
            formData.multi_end,
            time.time(),
        )
        encoder = hashlib.sha256()
        encoder.update(bytes(query_id, "utf-8"))
        query_id = encoder.hexdigest()

        dbPath = f"{self.static_path}/db/{formData.multi_database}"
        oligoDB = fp.query.OligoDatabase(dbPath)
        min_dist = oligoDB.get_oligo_min_dist()

        cmd = [
            "ifpd",
            "query",
            "set",
            shlex.quote(dbPath),
            shlex.quote(formData.multi_chromosome),
            shlex.quote(f"{self.static_path}/query/{query_id}"),
            shlex.quote(formData.multi_n_probes),
            "--order",
            shlex.quote(formData.f1),
            shlex.quote(formData.f2),
            shlex.quote(formData.f3),
            "--filter-thr",
            shlex.quote(formData.multi_f1_threshold),
            "--n-oligo",
            shlex.quote(formData.multi_n_oligo),
            "--min-d",
            shlex.quote(f"{min_dist}"),
            "--window-shift",
            shlex.quote(f"{formData.multi_win_shift}"),
        ]
        if len(queriedRegion) != 0:
            cmd.extend(queriedRegion)
        logging.info(" ".join(cmd))

        config = configparser.ConfigParser()
        timestamp = time.time()
        config["GENERAL"] = {
            "name": formData.multi_name,
            "description": formData.multi_description,
            "type": "spotting",
            "cmd": " ".join(cmd),
            "status": "queued",
        }
        config["WHEN"] = {
            "time": timestamp,
            "isotime": datetime.datetime.fromtimestamp(timestamp).isoformat(),
        }
        config["WHERE"] = {"db": formData.multi_database, "region": queriedRegion}
        config["WHAT"] = {
            "n_oligo": formData.multi_n_oligo,
            "threshold": formData.multi_f1_threshold,
            "n_probes": formData.multi_n_probes,
            "window_shift": formData.multi_win_shift,
        }
        config["HOW"] = {"f1": formData.f1, "f2": formData.f2, "f3": formData.f3}
        configPath = os.path.join(self.static_path, "query", f"{query_id}.config")
        with open(configPath, "w+") as OH:
            config.write(OH)

        self.queue.put(cmd)

        bot.response.status = 303
        bot.response.set_header(
            "Location", f"{self.root_uri}{self.app_uri}q/{query_id}"
        )

        # Output
        return "Query received."

    # Error --------------------------------------------------------------------

    def error500_with_redirect(routes, self, error):
        """Error 500.

        Args:
                self (App): ProbeDesigner.App instance.
                error: error data.
        """
        return f"""
        ERROR 500: nothing here, sorry :(<br/>
        Redirecting to homepage in 5 seconds.<br />
        <a href='{self.root_uri}{self.app_uri}'>Click here if you are stuck.</a>
        <meta http-equiv="refresh" content="5;url={self.root_uri}{self.app_uri}">
        """

    # AJAX Requests ------------------------------------------------------------

    def list_chromosomes(routes, self, dbDir):
        dbPath = os.path.join(self.static_path, "db", dbDir)
        chrList = [x for x in os.listdir(dbPath) if not os.path.isdir(x)]
        chrList = [x for x in chrList if x not in [".log", ".config"]]
        if not chrList:
            return '{"chrList":[]}'
        chrList.sort()
        return '{"chrList":["%s"]}' % '","'.join(chrList[::-1])

    def queueStatus(routes, self):
        taskList = []
        for task in self.queue.queue:
            query_id = os.path.basename(task[5])
            data = Query(query_id, self.qpath).data
            taskList.append(os.path.basename(task[1]) + f' @{data["isotime"]}')
        if not taskList:
            return '{"queue":[]}'
        return '{"queue": ["%s"]}' % '", "'.join(taskList)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ifpd',
 'ifpd.scripts',
 'ifpd.scripts.query',
 'ifpd.sections',
 'ifpd.sections.probe_design']

package_data = \
{'': ['*'],
 'ifpd': ['interface/css/*',
          'interface/fonts/*',
          'interface/images/*',
          'interface/js/*',
          'interface/views/*'],
 'ifpd.sections.probe_design': ['css/*', 'views/*']}

install_requires = \
['Paste>=3.5.0,<4.0.0',
 'bottle>=0.12.19,<0.13.0',
 'joblib>=1.0.1,<2.0.0',
 'matplotlib>=3.3.4,<4.0.0',
 'numpy>=1.20.1,<2.0.0',
 'pandas>=1.2.2,<2.0.0',
 'pytest>=6.2.2,<7.0.0',
 'rich>=9.10,<11.0',
 'scipy>=1.6.0,<2.0.0']

entry_points = \
{'console_scripts': ['ifpd = ifpd.scripts.ifpd:main']}

setup_kwargs = {
    'name': 'ifpd',
    'version': '2.1.1.post1',
    'description': 'An iFISH probe design pipeline, with web interface included.',
    'long_description': "# iFISH-Probe-Design\n\n[![DOI](https://zenodo.org/badge/143724120.svg)](https://zenodo.org/badge/latestdoi/143724120) ![](https://img.shields.io/github/license/ggirelli/ifish-probe-design.svg?style=flat)  \n![](https://github.com/ggirelli/ifish-probe-design/workflows/Python%20package/badge.svg?branch=main&event=push) ![PyPI - Python Version](https://img.shields.io/pypi/pyversions/ifpd) ![PyPI - Format](https://img.shields.io/pypi/format/ifpd) ![PyPI - Status](https://img.shields.io/pypi/status/ifpd)  \n![](https://img.shields.io/github/release/ggirelli/ifish-probe-design.svg?style=flat) ![](https://img.shields.io/github/release-date/ggirelli/ifish-probe-design.svg?style=flat) ![](https://img.shields.io/github/languages/code-size/ggirelli/ifish-probe-design.svg?style=flat)  \n![](https://img.shields.io/github/watchers/ggirelli/ifish-probe-design.svg?label=Watch&style=social) ![](https://img.shields.io/github/stars/ggirelli/ifish-probe-design.svg?style=social)\n\n[PyPi](https://pypi.org/project/ifpd/) | [docs](https://ggirelli.github.io/iFISH-probe-design/)\n\n\n**iFISH-Probe-Design** (`ifpd`) is a Python3.7+ package containing tools for selection of complementary oligonucleotides to build iFISH probes. It also includes a web interface, which simplifies the procedure by removing any requirement for programming skills. Read the online [documentation](https://ggirelli.github.io/iFISH-probe-design/) for more details.\n\n## Requirements\n\n**iFISH-Probe-Designer** is fully implemented in Python3.7+, thus you need Python3 to run it. Check out [here](https://realpython.com/installing-python/) how to install Python3.7+ on your machine if you don't have it yet.\n\n`ifpd` has been tested with Python 3.7, 3.8, and 3.9. We recommend installing it using `pipx` (see [below](https://github.com/ggirelli/ifish-probe-design#install)) to avoid dependency conflicts with other packages. The packages it depends on are listed in our [dependency graph](https://github.com/ggirelli/ifish-probe-design/network/dependencies). We use [`poetry`](https://github.com/python-poetry/poetry) to handle our dependencies.\n\n## Installation\n\nWe recommend installing `ifpd` using [`pipx`](https://github.com/pipxproject/pipx). Check how to install `pipx` [here](https://github.com/pipxproject/pipx#install-pipx) if you don't have it yet!\n\nOnce you have `pipx` ready on your system, install the latest stable release of `ifpd` by running: `pipx install ifpd`. If you see the stars (✨ 🌟 ✨), then the installation went well!\n\n## Usage\n\nAll **iFISH-Probe-Design** commands are accessible via the `ifpd` keyword on the terminal. For each command, you can access its help page by using the `-h` option. More details on how to run `ifpd` are available in the online [documentation](https://ggirelli.github.io/iFISH-probe-design/usage).\n\n## Contributing\n\nWe welcome any contributions to `ifpd`. In short, we use [`black`](https://github.com/psf/black) to standardize code format. Any code change also needs to pass `mypy` checks. For more details, please refer to our [contribution guidelines](https://github.com/ggirelli/ifish-probe-design/blob/main/CONTRIBUTING.md) if this is your first time contributing! Also, check out our [code of conduct](https://github.com/ggirelli/ifish-probe-design/blob/main/CODE_OF_CONDUCT.md).\n\n### Reference\n\n* Gelali, E., Girelli, G., Matsumoto, M., Wernersson, E., Custodio, J., Mota, A., ... & Bienko, M. (2019). iFISH is a publicly available resource enabling versatile DNA FISH to study genome architecture. Nature communications, 10(1), 1-15. ([link](https://www.nature.com/articles/s41467-019-09616-w))\n\n## License\n\n`MIT License - Copyright (c) 2016-21 Gabriele Girelli`\n",
    'author': 'Gabriele Girelli',
    'author_email': 'gigi.ga90@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ggirelli/iFISH-Probe-Design',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import configparser
import os


class Query(object):
    """Single query class."""

    def __init__(self, query_id, query_root):
        super(Query, self).__init__()
        assert Query.exists(query_id, query_root), f'query "{query_id}" not found.'

        self.data = {}
        self.data["id"] = query_id
        with open(os.path.join(query_root, f"{query_id}.config"), "r") as IH:
            config = configparser.ConfigParser()
            config.read_string("".join(IH.readlines()))
            self.data.update(config["GENERAL"].items())
            self.data.update(config["WHEN"].items())
            self.data.update(config["WHERE"].items())
            self.data.update(config["WHAT"].items())
            self.data.update(config["HOW"].items())

    @staticmethod
    def exists(query_id, query_root):
        """Check if a query exists."""
        return os.path.isfile(os.path.join(query_root, f"{query_id}.config"))



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import argparse
from ifpd import const, query
from ifpd.scripts import arguments as ap  # type: ignore
from ifpd.exception import enable_rich_assert
from joblib import Parallel, delayed  # type: ignore
import logging
import numpy as np  # type: ignore
import os
import pandas as pd  # type: ignore
from rich.logging import RichHandler  # type: ignore
from rich.progress import track  # type: ignore
import shutil

logging.basicConfig(
    level=logging.INFO,
    format="%(message)s",
    handlers=[RichHandler(markup=True, rich_tracebacks=True)],
)


def init_parser(subparsers: argparse._SubParsersAction) -> argparse.ArgumentParser:
    parser = subparsers.add_parser(
        __name__.split(".")[-1],
        description="""
Design a FISH probe set in a genomic region, with the aim of having the probes
as homogeneously spaced as possible. Concisely, the script does the following:
- Identify all probe candidates
- Calculate centrality, size, and homogeneity for each candidate
- Build a set of consecutive (nProbes+1) windows of the same size
    + Shift the windows set to build additional windows sets. Each windows set
      will produce one probe set candidate.
- For each window set:
    + Find the best probe in each window.
    + Aggregate each window's best probe into a candidate probe set.
- Rank candidate probe sets based on probe homogeneity.
- Return the top N candidates (maxSets), with plots, tables, fasta and bed.
""",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        help="Design a FISH probe set in a genomic region.",
    )
    parser.add_argument(
        "database", metavar="database", type=str, help="Path to database folder."
    )

    parser.add_argument(
        "chrom",
        type=str,
        help="Database feature to query for a probe set.",
    )
    parser.add_argument(
        "outdir",
        metavar="outDir",
        type=str,
        help="Path to query output directory. Stops if it exists already.",
    )
    parser.add_argument(
        "nProbes", metavar="nProbes", type=int, help="Number of probes to design."
    )

    parser.add_argument(
        "--region",
        type=int,
        nargs=2,
        default=(0, np.inf),
        help="""Start and end locations (space-separated) of the region of interest.
        When a region is not provided (or start/end coincide),
        the whole feature is queried.""",
    )
    parser.add_argument(
        "--n-oligo",
        metavar="nOligo",
        type=int,
        default=48,
        help="Number of oligos per probe. Default: 48",
    )
    parser.add_argument(
        "--max-sets",
        metavar="maxProbes",
        type=int,
        default=-1,
        help="""Maximum number of probe set candidates to output.
            Set to -1 to retrieve all candidates. Default: -1""",
    )
    parser = ap.add_version_option(parser)

    advanced = parser.add_argument_group("advanced arguments")
    advanced.add_argument(
        "--order",
        metavar="feature",
        type=str,
        default=const.featureList,
        nargs="+",
        help="""Space-separated features, used as explained in script description.
        The available features are: 'centrality', 'size', and 'homogeneity'. At least 2
        features must be listed. Default: "size homogeneity centrality".""",
    )
    advanced.add_argument(
        "--filter-thr",
        metavar="filterThr",
        type=float,
        default=0.1,
        help="""Threshold of first feature filter, used to identify
        a range around the best value (percentage range around it). Accepts values
        from 0 to 1. Default: 0.1""",
    )
    advanced.add_argument(
        "--min-d",
        metavar="minD",
        type=int,
        default=0,
        help="*DEPRECATED* Minimum distance between consecutive oligos. Default: 1",
    )
    advanced.add_argument(
        "--exact-n-oligo",
        action="store_const",
        dest="exact_n_oligo",
        const=True,
        default=False,
        help="""Stop if not enough oligos are found,
        instead of designing the largest probe.""",
    )
    advanced.add_argument(
        "--window-shift",
        metavar="winShift",
        type=float,
        default=0.1,
        help="""Window fraction for windows shifting.""",
    )
    advanced.add_argument(
        "-t",
        "--threads",
        metavar="nthreads",
        type=int,
        help="""Number of threads for parallelization. Default: 1""",
        default=1,
    )
    advanced.add_argument(
        "-f",
        action="store_const",
        dest="forceRun",
        const=True,
        default=False,
        help="""Force overwriting of the query if already run.
            This is potentially dangerous.""",
    )

    parser.set_defaults(parse=parse_arguments, run=run)

    return parser


def assert_region(args):
    if args.region is not None:
        if args.region[0] == args.region[1]:
            args.region = None
            return
        assert (
            args.region[0] >= 0
        ), f"start location cannot be negative [{args.region[0]}]."
        assert (
            args.region[1] >= 0
        ), f"end location cannot be negative [{args.region[1]}]."
        assert (
            args.region[1] > args.region[0]
        ), f"end location must be greater than start location [{args.region}]."


@enable_rich_assert
def parse_arguments(args: argparse.Namespace) -> argparse.Namespace:
    assert not os.path.isfile(
        args.outdir
    ), f"output folder expected, file found: {args.outdir}"
    if args.forceRun:
        if os.path.isdir(args.outdir):
            shutil.rmtree(args.outdir)
            logging.warning("Overwriting previously run query.")
    else:
        assert not os.path.isdir(
            args.outdir
        ), f"output folder already exists: {args.outdir}"

    assert_region(args)

    assert 2 <= len(
        args.order
    ), f"at least 2 features need, only {len(args.order)} found."
    for o in args.order:
        assert (
            o in const.featureList
        ), f'unrecognized feature "{o}". Should be one of {const.featureList}.'
    assert (
        args.filter_thr >= 0 and args.filter_thr <= 1
    ), f"first filter threshold must be a fraction: {args.filter_thr}"

    assert (
        args.window_shift > 0 and args.window_shift <= 1
    ), f"window shift must be a fraction: {args.window_shift}"

    assert (
        args.min_d >= 0
    ), f"negative minimum distance between consecutive oligos: {args.min_d}"
    assert args.n_oligo >= 1, f"a probe must have oligos: {args.n_oligo}"
    if args.max_sets == -1:
        args.max_sets = np.inf
    assert args.max_sets >= 0, f"at least 1 probe set in output: {args.max_sets}"

    return args


def init_db(
    args,
    oligoDB,
):
    assert (
        not oligoDB.has_overlaps()
    ), "databases with overlapping oligos are not supported yet."

    if args.chrom not in oligoDB.chromData.keys():
        oligoDB.read_chromosome(args.chrom)
    chromData = oligoDB.chromData[args.chrom]
    if args.region[1] == np.inf:
        args.region = (
            args.region[0],
            oligoDB.chromData[args.chrom]["chromEnd"].max(),
        )
    chromStart, chromEnd = args.region
    queried_region = (args.chrom, chromStart, chromEnd)

    selectCondition = np.logical_and(
        chromData.iloc[:, 0] >= chromStart, chromData.iloc[:, 1] <= chromEnd
    )
    selectedOligos = chromData.loc[selectCondition, :]

    return oligoDB, queried_region, selectCondition, selectedOligos


def build_candidates(args, queried_region, selectedOligos, oligoDB):
    logging.info("Build probe candidates.")
    args.threads = ap.check_threads(args.threads)
    if args.threads != 1:
        candidateList = Parallel(n_jobs=args.threads, backend="threading", verbose=1)(
            delayed(query.OligoProbe)(
                queried_region[0],
                selectedOligos.iloc[ii : (ii + args.n_oligo), :],
                oligoDB,
            )
            for ii in range(selectedOligos.shape[0] - args.n_oligo + 1)
        )

    else:
        candidateList = [
            query.OligoProbe(
                queried_region[0],
                selectedOligos.iloc[i : (i + args.n_oligo), :],
                oligoDB,
            )
            for i in track(range(selectedOligos.shape[0] - args.n_oligo + 1))
        ]
    logging.info(f"Found {len(candidateList)} probe candidates.")

    return candidateList


def build_windows(args, queried_region, oligoDB):
    chrom, chromStart, chromEnd = queried_region
    logging.info("Building window sets...")
    window_set = query.GenomicWindowList(None)
    window_size = chromEnd - chromStart
    window_size /= args.nProbes + 1
    window_size = int(window_size)

    skip = 1 + ((chromEnd - chromStart) % window_size != 0)
    for startPosition in range(chromStart, chromEnd, window_size)[:-skip]:
        window_set.add(chrom, startPosition, window_size)

    window_shift = max(
        int(args.window_shift * window_size),
        args.min_d + oligoDB.get_oligo_length_range()[0],
    )

    window_setList = [window_set.shift(s) for s in range(0, window_size, window_shift)]

    logging.info(f" Built {len(window_setList)} window sets.")

    return window_setList


def export_window_set(args, queried_region, window_setList, wsi):
    window_set = window_setList[wsi]
    window_set_path = os.path.join(args.outdir, f"probe_set_{wsi}")
    assert not os.path.isfile(window_set_path)
    assert not os.path.isdir(window_set_path)
    os.mkdir(window_set_path)

    fasta, bed = window_set.export(window_set_path, queried_region)
    fasta_path = os.path.join(window_set_path, f"probe_set_{wsi}.fa")
    bed_path = os.path.join(window_set_path, f"probe_set_{wsi}.bed")
    with open(fasta_path, "w+") as OH:
        OH.write(fasta)
    with open(bed_path, "w+") as OH:
        OH.write(bed)


def build_feature_table(args, queried_region, candidateList):
    logging.info("Describe candidates.")
    probeFeatureTable = query.ProbeFeatureTable(
        candidateList, queried_region, True, args.threads
    )

    logging.info("Write description table.")
    probeFeatureTable.data.to_csv(
        os.path.join(args.outdir, "probe_candidates.tsv"), "\t", index=False
    )

    assert args.nProbes <= probeFeatureTable.data.shape[0], "".join(
        [
            "not enough probes in the region of interest: ",
            f"{probeFeatureTable.data.shape[0]}/{args.nProbes}",
        ]
    )

    return probeFeatureTable


def populate_windows(args, candidateList, window_setList, probeFeatureTable):
    for wsi in range(len(window_setList)):
        window_set = window_setList[wsi]
        for wi in range(len(window_set)):
            window = window_set[wi]
            probeFeatureTable.reset()

            selectCondition = np.logical_and(
                probeFeatureTable.data.loc[:, "chromStart"] >= window.chromStart,
                probeFeatureTable.data.loc[:, "chromEnd"] <= window.chromEnd,
            )
            logging.info(
                "".join(
                    [
                        f"Found {selectCondition.sum()} probe candidates",
                        f" in window #{wi} of set #{wsi}.",
                    ]
                )
            )

            if selectCondition.sum() == 0:
                window_setList[wsi][wi].probe = None
                continue
            elif selectCondition.sum() != 1:
                probeFeatureTable.keep(selectCondition, cumulative=True)
                feature_range, feature = probeFeatureTable.filter(
                    args.order[0], args.filter_thr, cumulative=True
                )
                feature_range = np.round(feature_range, 6)
                logging.info(
                    "".join(
                        [
                            f" Selected {probeFeatureTable.data.shape[0]}",
                            f" candidates in the range {feature_range} of '{feature}'.",
                        ]
                    )
                )
                logging.info(
                    " Ranking probe candidates based on" + f" '{args.order[1]}'."
                )
                probeFeatureTable.rank(args.order[1])

            window_setList[wsi][wi].probe = candidateList[
                probeFeatureTable.data.index[0]
            ]
    return window_setList


@enable_rich_assert
def run(args: argparse.Namespace) -> None:
    os.mkdir(args.outdir)
    ap.add_log_file_handler(os.path.join(args.outdir, "log"))

    logging.info("Read database.")
    oligoDB = query.OligoDatabase(args.database)
    oligoDB, queried_region, selectCondition, selectedOligos = init_db(args, oligoDB)
    args = ap.check_n_oligo(args, selectCondition)

    candidateList = build_candidates(args, queried_region, selectedOligos, oligoDB)
    probeFeatureTable = build_feature_table(args, queried_region, candidateList)
    window_setList = build_windows(args, queried_region, oligoDB)
    window_setList = populate_windows(
        args, candidateList, window_setList, probeFeatureTable
    )

    logging.info("Compare probe set candidates.")
    probeSetSpread = np.array(
        [ws.calc_probe_size_and_homogeneity() for ws in window_setList]
    )
    probeCount = np.array([ws.count_probes() for ws in window_setList])
    logging.info(
        "".join(
            [
                f" {sum(probeCount == args.nProbes)}/{len(window_setList)}",
                f" probe set candidates have {args.nProbes} probes.",
            ]
        )
    )

    logging.info("Rank based on #probes and homogeneity (of probes and size).")
    probeSetData = pd.DataFrame.from_dict(
        {
            "id": range(len(window_setList)),
            "homogeneity": probeSetSpread[np.argsort(probeCount)[::-1]],
            "nProbes": probeCount[np.argsort(probeCount)[::-1]],
        }
    )
    probeSetData.sort_values(
        by=["nProbes", "homogeneity"], ascending=[False, False], inplace=True
    )
    probeSetData.drop("id", axis=1).to_csv(
        os.path.join(args.outdir, "set_candidates.tsv"), "\t", index=False
    )

    logging.info("Export probe set candidates.")
    window_setList = [window_setList[i] for i in probeSetData["id"].values]

    if args.threads != 1:
        Parallel(n_jobs=args.threads, verbose=1)(
            delayed(export_window_set)(args, queried_region, window_setList, wsi)
            for wsi in range(len(window_setList))
        )
    else:
        for wsi in track(range(len(window_setList))):
            export_window_set(args, queried_region, window_setList, wsi)

    logging.info("Done. :thumbs_up: :smiley:")
